import numpy as np
import pandas as pd

from fedot.core.data.data import InputData, data_type_is_table, data_type_is_ts, data_type_is_multi_ts
from fedot.core.repository.dataset_types import DataTypesEnum


def data_type_is_suitable_preprocessing(data: InputData) -> bool:
    if data_type_is_table(data) or data_type_is_ts(data) or data_type_is_multi_ts(data):
        return True
    return False


def replace_inf_with_nans(input_data: InputData):
    values_to_replace = [np.inf, -np.inf]
    features_with_replaced_inf = np.where(np.isin(input_data.features,
                                                  values_to_replace),
                                          np.nan,
                                          input_data.features)
    input_data.features = features_with_replaced_inf


def replace_nans_with_empty_strings(input_data: InputData):
    """
    Replace NaNs with empty strings in input_data.features
    """
    input_data.features = np.where(pd.isna(input_data.features),
                                   '',
                                   input_data.features)


def convert_into_column(array: np.array):
    """ Perform conversion for data if it is necessary """
    if len(array.shape) == 1:
        return array.reshape(-1, 1)
    else:
        return array


def divide_data_categorical_numerical(input_data: InputData, categorical_ids: list,
                                      non_categorical_ids: list) -> (InputData, InputData):
    """
    Split tabular InputData into two parts: with numerical and categorical features
    using list with ids of categorical and numerical features.
    """

    if len(categorical_ids) > 0 and len(non_categorical_ids) > 0:
        # Both categorical and numerical features
        numerical_input = input_data.subset_features(non_categorical_ids)
        categorical_input = input_data.subset_features(categorical_ids)
        return numerical_input, categorical_input

    elif len(categorical_ids) == 0 and len(non_categorical_ids) > 0:
        # Only numerical
        numerical_input = input_data.subset_features(non_categorical_ids)
        return numerical_input, None

    elif len(categorical_ids) > 0 and len(non_categorical_ids) == 0:
        # Only categorical
        categorical_input = input_data.subset_features(categorical_ids)
        return None, categorical_input

    else:
        prefix = 'InputData contains no categorical and no numerical features.'
        raise ValueError(f'{prefix} Check data for Nans and inf values')


def find_categorical_columns(table: np.array, column_types: dict = None):
    """
    Method for finding categorical and non-categorical columns in tabular data

    :param table: tabular data for string columns types determination
    :param column_types: list with column types. If None, perform default checking
    :return categorical_ids: indices of categorical columns in table
    :return non_categorical_ids: indices of non categorical columns in table
    """
    if column_types is None:
        # Define if data contains string columns for "unknown table"
        return force_categorical_determination(table)

    categorical_ids = []
    non_categorical_ids = []
    for column_id, type_name in enumerate(column_types):
        if 'str' in str(type_name):
            categorical_ids.append(column_id)
        else:
            non_categorical_ids.append(column_id)

    return categorical_ids, non_categorical_ids


def force_categorical_determination(table):
    """ Find string columns using 'computationally expensive' approach """
    source_shape = table.shape
    columns_number = source_shape[1] if len(source_shape) > 1 else 1

    categorical_ids = []
    non_categorical_ids = []
    # For every column in table make check for first element
    for column_id in range(0, columns_number):
        column = table[:, column_id] if columns_number > 1 else table
        col_shape = column.shape
        for i in column:
            # Check if element is string object or not until the first appearance
            if len(col_shape) == 2 and isinstance(i[0], str):
                # Column looks like [[n], [n], [n]]
                categorical_ids.append(column_id)
                break
            elif len(col_shape) == 1 and isinstance(i, str):
                # Column [n, n, n]
                categorical_ids.append(column_id)
                break

        if column_id not in categorical_ids:
            non_categorical_ids.append(column_id)

    return categorical_ids, non_categorical_ids


def data_has_missing_values(data: InputData) -> bool:
    """ Check data for missing values."""
    if data_type_is_suitable_preprocessing(data):
        return pd.DataFrame(data.features).isna().sum().sum() > 0
    return False


def data_has_categorical_features(data: InputData) -> bool:
    """
    Check data for categorical columns.
    Return bool, whether data has categorical columns or not
    """
    if data.data_type is not DataTypesEnum.table:
        return False

    features_types = data.supplementary_data.column_types.get('features')
    cat_ids, non_cat_ids = find_categorical_columns(data.features, features_types)
    data_has_categorical_columns = len(cat_ids) > 0

    return data_has_categorical_columns


def data_has_text_features(data: InputData) -> bool:
    """
    Checks data for text fields.
    Data with text fields is always 1-dimensional due to the previous
    parsing of it from a general table to the distinct text data source.
    Returns bool, whether data has text fields or not
    """
    # TODO andreygetmanov: make compatible with current text checking
    if data.data_type is DataTypesEnum.text:
        return True
    return False



from copy import copy
from typing import Optional

from fedot.api.time import ApiTime
from fedot.core.constants import BEST_QUALITY_PRESET_NAME, \
    FAST_TRAIN_PRESET_NAME
from fedot.core.repository.dataset_types import DataTypesEnum
from fedot.core.repository.operation_types_repository import OperationTypesRepository, get_operations_for_task
from fedot.core.repository.tasks import Task


class OperationsPreset:
    """ Class for presets processing. Preset is a set of operations (data operations
    and models), which will be used during pipeline structure search
    """

    def __init__(self, task: Task, preset_name: str):
        self.task = task
        self.preset_name = preset_name

        # Is there a modification in preset or not
        self.modification_using = False

    def composer_params_based_on_preset(self, api_params: dict, data_type: Optional[DataTypesEnum] = None) -> dict:
        """ Return composer parameters dictionary with appropriate operations
        based on defined preset
        """
        updated_params = copy(api_params)

        if self.preset_name is None and 'preset' in updated_params:
            self.preset_name = updated_params['preset']

        if self.preset_name is not None and 'available_operations' not in api_params:
            available_operations = self.filter_operations_by_preset(data_type)
            updated_params['available_operations'] = available_operations

        return updated_params

    def filter_operations_by_preset(self, data_type: Optional[DataTypesEnum] = None):
        """ Filter operations by preset, remove "heavy" operations and save
        appropriate ones
        """
        preset_name = self.preset_name
        if 'auto' in preset_name:
            available_operations = get_operations_for_task(self.task, data_type, mode='all')
            return available_operations

        # TODO remove workaround
        # Use best_quality preset but exclude several operations
        if 'stable' in self.preset_name:
            # Use best_quality preset but exclude several operations
            preset_name = BEST_QUALITY_PRESET_NAME
        excluded = ['mlp', 'svc', 'svr', 'arima', 'exog_ts', 'text_clean',
                    'catboost', 'lda', 'qda', 'lgbm', 'one_hot_encoding',
                    'resample', 'stl_arima']
        excluded_tree = ['xgboost', 'catboost', 'xgbreg', 'catboostreg']

        if '*' in preset_name:
            self.modification_using = True
            # The modification has been added
            preset_name, modification = preset_name.split('*')
            modification = ''.join(('*', modification))

            mod_operations = get_operations_for_task(self.task, data_type, mode='all', preset=modification)

        # Get operations
        available_operations = get_operations_for_task(self.task, data_type, mode='all', preset=preset_name)

        if self.modification_using:
            # Find subsample of operations
            filtered_operations = set(available_operations).intersection(set(mod_operations))
            available_operations = list(filtered_operations)

        # Exclude "heavy" operations if necessary
        if 'stable' in self.preset_name:
            available_operations = self.new_operations_without_heavy(excluded, available_operations)

        if 'gpu' in self.preset_name:
            repository = OperationTypesRepository().assign_repo('model', 'gpu_models_repository.json')
            available_operations = repository.suitable_operation(task_type=self.task.task_type, data_type=data_type)

        filtered_operations = set(available_operations).difference(set(excluded_tree))
        available_operations = list(filtered_operations)

        return available_operations

    @staticmethod
    def new_operations_without_heavy(excluded_operations, available_operations) -> list:
        """ Create new list without heavy operations """
        available_operations = [_ for _ in available_operations if _ not in excluded_operations]

        return available_operations


def change_preset_based_on_initial_fit(timer: ApiTime, n_jobs: int) -> str:
    """
    If preset was set as 'auto', based on initial pipeline fit time, appropriate one can be chosen
    """
    if timer.time_for_automl in [-1, None]:
        return BEST_QUALITY_PRESET_NAME

    # Change preset to appropriate one

    if timer.have_time_for_the_best_quality(n_jobs=n_jobs):
        # It is possible to train only few number of pipelines during optimization - use simplified preset
        return BEST_QUALITY_PRESET_NAME
    else:
        return FAST_TRAIN_PRESET_NAME



import platform
from multiprocessing import set_start_method
from pathlib import Path
from typing import Dict, List, Optional, Sequence, Type, Union

from fedot.core.caching.pipelines_cache import OperationsCache
from fedot.core.caching.preprocessing_cache import PreprocessingCache
from fedot.core.composer.composer import Composer
from fedot.core.composer.gp_composer.gp_composer import GPComposer
from fedot.core.log import LoggerAdapter, default_log
from fedot.core.optimisers.gp_comp.gp_optimizer import EvoGraphOptimizer
from fedot.core.optimisers.gp_comp.gp_params import GPGraphOptimizerParameters
from fedot.core.optimisers.gp_comp.pipeline_composer_requirements import PipelineComposerRequirements
from fedot.core.optimisers.initial_graphs_generator import InitialPopulationGenerator, GenerationFunction
from fedot.core.optimisers.objective.metrics_objective import MetricsObjective
from fedot.core.optimisers.optimizer import GraphOptimizer, GraphOptimizerParameters, GraphGenerationParams
from fedot.core.pipelines.pipeline import Pipeline
from fedot.core.pipelines.pipeline_graph_generation_params import get_pipeline_generation_params
from fedot.core.pipelines.verification import rules_by_task
from fedot.core.repository.operation_types_repository import get_operations_for_task
from fedot.core.repository.quality_metrics_repository import (
    ComplexityMetricsEnum,
    MetricsEnum,
    MetricType
)
from fedot.core.repository.tasks import Task
from fedot.core.utilities.data_structures import ensure_wrapped_in_sequence
from fedot.remote.remote_evaluator import RemoteEvaluator
from fedot.utilities.define_metric_by_task import MetricByTask


def set_multiprocess_start_method():
    system = platform.system()
    if system == 'Linux':
        set_start_method("spawn", force=True)


class ComposerBuilder:

    def __init__(self, task: Task):
        self.log: LoggerAdapter = default_log(self)

        self.task: Task = task
        self.metrics: Sequence[MetricsEnum] = MetricByTask.get_default_quality_metrics(task.task_type)

        self.optimizer_cls: Type[GraphOptimizer] = EvoGraphOptimizer  # default optimizer class
        self.optimizer_parameters: Optional[GraphOptimizerParameters] = None
        self.optimizer_external_parameters: dict = {}

        self.composer_cls: Type[Composer] = GPComposer  # default composer class
        self.composer_requirements: Optional[PipelineComposerRequirements] = None
        self.graph_generation_params: Optional[GraphGenerationParams] = None

        self.initial_population: Union[Pipeline, Sequence[Pipeline]] = ()
        self.initial_population_generation_function: Optional[GenerationFunction] = None

        self._keep_history: bool = True
        self._full_history_dir: Optional[Path] = None

        self.pipelines_cache: Optional[OperationsCache] = None
        self.preprocessing_cache: Optional[PreprocessingCache] = None

    def with_composer(self, composer_cls: Optional[Type[Composer]]):
        if composer_cls is not None:
            self.composer_cls = composer_cls
        return self

    def with_optimizer(self, optimizer_cls: Optional[Type[GraphOptimizer]]):
        if optimizer_cls:
            self.optimizer_cls = optimizer_cls
        return self

    def with_optimizer_params(self, parameters: Optional[GraphOptimizerParameters] = None,
                              external_parameters: Optional[Dict] = None):
        if parameters is not None:
            self.optimizer_parameters = parameters
        if external_parameters is not None:
            self.optimizer_external_parameters = external_parameters
        return self

    def with_requirements(self, requirements: PipelineComposerRequirements):
        self.composer_requirements = requirements
        if self.composer_requirements.max_pipeline_fit_time:
            set_multiprocess_start_method()
        return self

    def with_graph_generation_param(self, graph_generation_params: GraphGenerationParams):
        self.graph_generation_params = graph_generation_params
        return self

    def with_metrics(self, metrics: Union[MetricType, List[MetricType]]):
        self.metrics = ensure_wrapped_in_sequence(metrics)
        return self

    def with_initial_pipelines(self, initial_pipelines: Union[Pipeline, Sequence[Pipeline]]):
        self.initial_population = initial_pipelines
        return self

    def with_initial_pipelines_generation_function(self, generation_function: GenerationFunction):
        self.initial_population_generation_function = generation_function
        return self

    def with_cache(self, pipelines_cache: Optional[OperationsCache] = None,
                   preprocessing_cache: Optional[PreprocessingCache] = None):
        self.pipelines_cache = pipelines_cache
        self.preprocessing_cache = preprocessing_cache
        return self

    @staticmethod
    def _get_default_composer_params(task: Task) -> PipelineComposerRequirements:
        # Get all available operations for task
        operations = get_operations_for_task(task=task, mode='all')
        return PipelineComposerRequirements(primary=operations, secondary=operations)

    def _get_default_graph_generation_params(self) -> GraphGenerationParams:
        return get_pipeline_generation_params(
            rules_for_constraint=rules_by_task(self.task.task_type),
            task=self.task,
            requirements=self.composer_requirements
        )

    @staticmethod
    def _get_default_complexity_metrics() -> List[MetricsEnum]:
        return [ComplexityMetricsEnum.node_num]

    def build(self) -> Composer:
        multi_objective = len(self.metrics) > 1
        if not self.composer_requirements:
            self.composer_requirements = self._get_default_composer_params(self.task)
        if not self.graph_generation_params:
            self.graph_generation_params = self._get_default_graph_generation_params()
        if not self.optimizer_parameters:
            self.optimizer_parameters = GPGraphOptimizerParameters(multi_objective=multi_objective)
        if not multi_objective:
            # Add default complexity metric for supplementary comparison of individuals with equal fitness
            self.metrics = self.metrics + self._get_default_complexity_metrics()
        if RemoteEvaluator().is_enabled:
            # This explicit passing of singleton evaluator isolates optimizer
            #  from knowing about it. Needed for separation of FEDOT from GOLEM.
            #  Best solution would be to avoid singleton init & access altogether,
            #  instead passing remote evaluator & its params explicitly through API.
            self.graph_generation_params.remote_evaluator = RemoteEvaluator()

        objective = MetricsObjective(self.metrics, multi_objective)

        initial_population = InitialPopulationGenerator(self.optimizer_parameters.pop_size,
                                                        self.graph_generation_params,
                                                        self.composer_requirements) \
            .with_initial_graphs(self.initial_population) \
            .with_custom_generation_function(self.initial_population_generation_function)()

        optimiser = self.optimizer_cls(objective=objective,
                                       initial_graphs=initial_population,
                                       requirements=self.composer_requirements,
                                       graph_generation_params=self.graph_generation_params,
                                       graph_optimizer_params=self.optimizer_parameters,
                                       **self.optimizer_external_parameters)

        composer = self.composer_cls(optimiser,
                                     self.composer_requirements,
                                     self.pipelines_cache,
                                     self.preprocessing_cache)

        return composer



from functools import reduce
from typing import Optional

import numpy as np


def find_common_elements(*arrays: np.array) -> np.array:
    """
    Intersects all of the ndarrays in ``arrays`` and returns their common elements

    :param arrays: tuple of ndarrays to be intersected

    :return: ndarray of unique elements common to all the provided ``arrays``
        or empty ndarray if none of them intersects
    """
    common_elements = reduce(np.intersect1d, arrays[1:], arrays[0])
    return common_elements


def flatten_extra_dim(data: Optional[np.array]) -> Optional[np.array]:
    """
    Removes last dimension if it is equal to one

    :param data: ndarray to be reshaped

    :return: reshaped view of the original array or None if input is None
    """

    if data is not None and data.shape[-1] == 1:
        return np.squeeze(data, axis=-1)
    return data


def atleast_n_dimensions(data: np.array, ndim: int) -> np.array:
    """
    Returns a view of the ``data` with at least ``ndim`` dimensions

    :param data: ndarray which dimensional size should be set to at least ``ndim``
    :param ndim: number of required axes to have in ``data``

    :return: ``data`` expanded from the last axis to the provided ``ndim`` size if it doesn't satisfy it
    """
    while data.ndim < ndim:
        data = np.expand_dims(data, axis=-1)
    return data


def atleast_2d(data: np.array) -> np.array:
    """
    Returns a view of the ``data` with at least `2` dimensions

    :param data: ndarray which dimensional size should be set to at least 2

    :return: ``data`` expanded from the last axis to `2` dimensions size if it doesn't satisfy it
    """
    return atleast_n_dimensions(data, ndim=2)


def atleast_4d(data: np.array) -> np.array:
    """
    Returns a view of the ``data` with at least `4` dimensions

    :param data: ndarray which dimensional size should be set to at least 4

    :return: ``data`` expanded from the last axis to `4` dimensions size if it doesn't satisfy it
    """
    return atleast_n_dimensions(data, ndim=4)



from typing import Any, Dict, Type, Tuple

from fedot.core.optimisers.objective import Objective
from fedot.core.optimisers.objective.metrics_objective import MetricsObjective
from fedot.core.serializers.any_serialization import any_from_json


def objective_from_json(cls: Type[Objective], json_obj: Dict[str, Any]) -> Objective:
    # backward compatibility
    metrics = json_obj.get('metrics')
    if metrics and not isinstance(metrics[0], Tuple):
        return any_from_json(MetricsObjective, json_obj)
    return any_from_json(cls, json_obj)



import os

import numpy as np
from matplotlib import pyplot as plt
from sklearn.metrics import mean_absolute_error, mean_squared_error, mean_absolute_percentage_error

from examples.simple.time_series_forecasting.ts_pipelines import ts_complex_ridge_smoothing_pipeline
from fedot.api.main import Fedot
from fedot.core.data.data import InputData
from fedot.core.data.data_split import train_test_data_setup
from fedot.core.repository.tasks import TsForecastingParams, Task, TaskTypesEnum
from fedot.core.utils import fedot_project_root


def prepare_data(forecast_length, is_multi_ts):
    """
    Function to form InputData from file with time-series
    """
    columns_to_use = ['61_91', '56_86', '61_86', '66_86']
    target_column = '61_91'
    task = Task(TaskTypesEnum.ts_forecasting,
                TsForecastingParams(forecast_length=forecast_length))
    file_path = os.path.join(str(fedot_project_root()), 'cases/data/arctic/topaz_multi_ts.csv')
    if is_multi_ts:
        data = InputData.from_csv_multi_time_series(
            file_path=file_path,
            task=task,
            columns_to_use=columns_to_use)
    else:
        data = InputData.from_csv_time_series(
            file_path=file_path,
            task=task,
            target_column=target_column)
    train_data, test_data = train_test_data_setup(data)
    return train_data, test_data, task


def run_multi_ts_forecast(forecast_length, is_multi_ts):
    """
    Function for run experiment with use multi_ts data type (is_multi_ts=True) for train set extension
    Or run experiment on one time-series (is_multi_ts=False)
    """
    train_data, test_data, task = prepare_data(forecast_length, is_multi_ts)
    # init model for the time series forecasting
    init_pipeline = ts_complex_ridge_smoothing_pipeline()
    model = Fedot(problem='ts_forecasting',
                  task_params=task.task_params,
                  timeout=5,
                  n_jobs=1,
                  max_depth=5,
                  num_of_generations=20,
                  pop_size=15,
                  max_arity=4,
                  cv_folds=None,
                  validation_blocks=None,
                  initial_assumption=init_pipeline
                  )
    # fit model
    pipeline = model.fit(train_data)
    pipeline.show()

    # use model to obtain forecast
    forecast = model.predict(test_data)
    target = np.ravel(test_data.target)

    # visualize results
    model.plot_prediction()

    print(f'MAE: {mean_absolute_error(target, forecast)}')
    print(f'RMSE: {mean_squared_error(target, forecast)}')
    print(f'MAPE: {mean_absolute_percentage_error(target, forecast)}')

    print(model.get_metrics(metric_names=['rmse', 'mae', 'mape'], target=target))


if __name__ == '__main__':
    forecast_length = 60
    run_multi_ts_forecast(forecast_length, is_multi_ts=True)
    run_multi_ts_forecast(forecast_length, is_multi_ts=False)



from fedot.api.main import Fedot
from fedot.core.utils import fedot_project_root


def run_classification_example(timeout: float = None, visualization=False):
    problem = 'classification'
    train_data_path = f'{fedot_project_root()}/cases/data/scoring/scoring_train.csv'
    test_data_path = f'{fedot_project_root()}/cases/data/scoring/scoring_test.csv'

    baseline_model = Fedot(problem=problem, timeout=timeout, seed=42)
    baseline_model.fit(features=train_data_path, target='target', predefined_model='rf')

    baseline_model.predict(features=test_data_path)
    print(baseline_model.get_metrics())

    auto_model = Fedot(problem=problem, seed=42, timeout=timeout, n_jobs=-1, preset='best_quality',
                       max_pipeline_fit_time=5, metric='roc_auc')
    auto_model.fit(features=train_data_path, target='target')
    prediction = auto_model.predict_proba(features=test_data_path)
    print(auto_model.get_metrics())
    if visualization:
        auto_model.plot_prediction()
    return prediction


if __name__ == '__main__':
    run_classification_example(timeout=10.0, visualization=True)



from copy import deepcopy
import numpy as np
import pandas as pd

from sklearn.preprocessing import LabelEncoder
from fedot.core.data.data import InputData
from fedot.core.data.data_preprocessing import find_categorical_columns
from fedot.preprocessing.data_types import NAME_CLASS_INT, FEDOT_STR_NAN


class BinaryCategoricalPreprocessor:
    """ Class for categories features preprocessing: converting binary string features into integers. """

    def __init__(self):
        self.binary_encoders = {}
        self.binary_ids_to_convert = []

        # List with binary categorical features ids which contain Nans
        self.binary_features_with_nans = []

    def fit(self, input_data: InputData):
        """
        Find indices of columns which are contains categorical values. Binary features and at the same time
        has str objects. If there are such features - convert it into int
        """
        features_types = input_data.supplementary_data.column_types['features']
        categorical_ids, non_categorical_ids = find_categorical_columns(table=input_data.features,
                                                                        column_types=features_types)
        if len(categorical_ids) == 0:
            # There is no need to process categorical features
            return self

        binary_ids_to_convert = []
        number_of_columns = input_data.features.shape[-1]
        for column_id, number in enumerate(range(number_of_columns)):
            column = np.array(input_data.features[:, column_id])

            # Numpy with strings cannot be processed for nans search - so use pandas
            pd_column = pd.Series(column)
            is_row_has_nan = pd.isna(pd_column)
            nans_number = is_row_has_nan.sum()
            if nans_number > 0 and column_id in categorical_ids:
                # This categorical column has nans
                column, gap_ids = replace_nans_with_fedot_nans(column, is_row_has_nan)
                column_uniques = np.unique(column)

                if len(column_uniques) <= 3:
                    # There is column with binary categories and gaps
                    self.binary_features_with_nans.append(column_id)
                    binary_ids_to_convert.append(column_id)
                    self._train_encoder(column, column_id)
            else:
                column_uniques = np.unique(column)
                if len(column_uniques) <= 2 and column_id in categorical_ids:
                    # Column contains binary string feature
                    binary_ids_to_convert.append(column_id)

                    # Train encoder for current column
                    self._train_encoder(column, column_id)

        self.binary_ids_to_convert = binary_ids_to_convert
        return self

    def transform(self, input_data: InputData) -> InputData:
        """
        Apply transformation (converting str into integers) for selected (while training) features.
        """
        if len(self.binary_ids_to_convert) == 0:
            # There are no binary categorical features
            return input_data

        converted_features = []
        number_of_columns = input_data.features.shape[-1]
        for column_id, number in enumerate(range(number_of_columns)):
            if column_id in self.binary_ids_to_convert:
                # If column contains nans - replace them with fedot nans special string
                column = input_data.features[:, column_id]
                is_row_has_nan = pd.isna(pd.Series(column))
                column, gap_ids = replace_nans_with_fedot_nans(column, is_row_has_nan)

                # Convert into integers
                converted_column = self._apply_encoder(column, column_id, gap_ids)
            else:
                # Stay column the same
                converted_column = np.array(input_data.features[:, column_id])

            converted_features.append(converted_column.reshape((-1, 1)))

        # Store transformed features
        copied_data = deepcopy(input_data)
        copied_data.features = np.hstack(converted_features)

        # Update features types
        features_types = copied_data.supplementary_data.column_types['features']
        for converted_column_id in self.binary_ids_to_convert:
            features_types[converted_column_id] = NAME_CLASS_INT
        return copied_data

    def _train_encoder(self, column: np.array, column_id: int):
        """ Convert labels in the column from string into int via Label encoding.
        So, Label encoder is fitted to do such transformation.
        """
        encoder = LabelEncoder()
        encoder.fit(column)

        # Store fitted label encoder for transform method
        self.binary_encoders.update({column_id: encoder})

    def _apply_encoder(self, column: np.array, column_id: int, gap_ids: np.array) -> np.array:
        """ Apply already fitted encoders """
        encoder = self.binary_encoders[column_id]
        encoder_classes = list(encoder.classes_)

        # If the column contains categories not previously encountered
        for label in list(set(column)):
            if label not in encoder_classes:
                encoder_classes.append(label)

        # Extent encoder classes
        encoder.classes_ = np.array(encoder_classes)

        converted = encoder.transform(column)
        if len(gap_ids) > 0:
            # Column has nans in its structure - after conversion replace it
            converted = converted.astype(float)
            converted[gap_ids] = np.nan

        return converted


def replace_nans_with_fedot_nans(column: np.array, is_row_has_nan):
    # There are nans in the columns - find indices of such objects
    # True > 0
    gap_ids = np.ravel(np.argwhere(is_row_has_nan.values > 0))

    # Add new category - 'fedot_nan' after converting it will be replaced by nans
    column[gap_ids] = FEDOT_STR_NAN
    return column, gap_ids



from abc import abstractmethod, ABC
from random import choice
from typing import Optional, Iterable

from fedot.core.optimisers.gp_comp.pipeline_composer_requirements import PipelineComposerRequirements
from fedot.core.optimisers.graph import OptNode


class OptNodeFactory(ABC):
    @abstractmethod
    def exchange_node(self,
                      node: OptNode) -> Optional[OptNode]:
        """
        Returns new node based on a current node using information about node and advisor.

        :param node: current node that must be changed.
        """
        pass

    @abstractmethod
    def get_parent_node(self, node: OptNode, **kwargs) -> Optional[OptNode]:
        """
        Returns new parent node for the current node
        based on the content of the current node and using advisor.

        :param node: current node for which a parent node is generated
        """
        pass

    @abstractmethod
    def get_node(self, **kwargs) -> Optional[OptNode]:
        """
        Returns new node based on the requirements for a node.
        """
        pass


class DefaultOptNodeFactory(OptNodeFactory):
    def __init__(self, available_node_types: Iterable[str]):
        self._available_nodes = tuple(available_node_types)

    def exchange_node(self, node: OptNode) -> OptNode:
        return node

    def get_parent_node(self, node: OptNode, **kwargs) -> OptNode:
        return self.get_node(**kwargs)

    def get_node(self, **kwargs) -> OptNode:
        return OptNode(content={'name': choice(self._available_nodes)})



from typing import List

from SALib.analyze.sobol import analyze as sobol_analyze
from SALib.sample import saltelli

from fedot.core.pipelines.pipeline import Pipeline
import numpy as np


def sobol_method(problem: dict, samples: List[Pipeline], operation_response) -> dict:
    indices = sobol_analyze(problem, operation_response, print_to_console=False)

    return indices


def make_saltelly_sample(problem: dict, num_of_samples=100) -> np.array:
    samples = saltelli.sample(problem, num_of_samples)

    return samples


analyze_method_by_name = {
    'sobol': sobol_method,
}

sample_method_by_name = {
    'saltelli': make_saltelly_sample

}



from fedot.core.pipelines.pipeline_builder import PipelineBuilder


def ts_ets_pipeline():
    """
    Return pipeline with the following structure:

    .. image:: img_ts_pipelines/ts_ets_pipeline.png
      :width: 55%

    Where cut - cut part of dataset and ets - exponential smoothing
    """
    pip_builder = PipelineBuilder().add_node('cut').add_node('ets',
                                                             params={'error': 'add',
                                                                     'trend': 'add',
                                                                     'seasonal': 'add',
                                                                     'damped_trend': False,
                                                                     'seasonal_periods': 20})
    pipeline = pip_builder.to_pipeline()
    return pipeline


def ts_ets_ridge_pipeline():
    """
    Return pipeline with the following structure:

    .. image:: img_ts_pipelines/ts_ets_ridge_pipeline.png
      :width: 55%

    Where cut - cut part of dataset, ets - exponential smoothing
   """
    pip_builder = PipelineBuilder()\
        .add_sequence(('cut', {'cut_part': 0.5}),
                      ('ets', {'error': 'add', 'trend': 'add', 'seasonal': 'add',
                               'damped_trend': False, 'seasonal_periods': 20}),
                      branch_idx=0)\
        .add_sequence('lagged', 'ridge', branch_idx=1).join_branches('ridge')

    pipeline = pip_builder.to_pipeline()
    return pipeline


def ts_glm_pipeline():
    """
    Return pipeline with the following structure:

    .. image:: img_ts_pipelines/ts_glm_pipeline.png
      :width: 55%

    Where glm - Generalized linear model
    """
    pipeline = PipelineBuilder().add_node('glm', params={'family': 'gaussian'}).to_pipeline()
    return pipeline


def ts_glm_ridge_pipeline():
    """
    Return pipeline with the following structure:

    .. image:: img_ts_pipelines/ts_glm_ridge_pipeline.png
      :width: 55%

    Where glm - Generalized linear model
    """
    pip_builder = PipelineBuilder() \
        .add_sequence('glm', branch_idx=0) \
        .add_sequence('lagged', 'ridge', branch_idx=1).join_branches('ridge')

    pipeline = pip_builder.to_pipeline()
    return pipeline


def ts_polyfit_pipeline(degree):
    """
    Return pipeline with the following structure:

    .. image:: img_ts_pipelines/ts_polyfit_pipeline.png
      :width: 55%

    Where polyfit - Polynomial interpolation
    """
    pipeline = PipelineBuilder().add_node('polyfit', params={'degree': degree}).to_pipeline()
    return pipeline


def ts_polyfit_ridge_pipeline(degree):
    """
    Return pipeline with the following structure:

    .. image:: img_ts_pipelines/ts_polyfit_ridge_pipeline.png
      :width: 55%

    Where polyfit - Polynomial interpolation
    """
    pip_builder = PipelineBuilder() \
        .add_sequence(('polyfit', {'degree': degree}), branch_idx=0) \
        .add_sequence('lagged', 'ridge', branch_idx=1).join_branches('ridge')

    pipeline = pip_builder.to_pipeline()
    return pipeline


def ts_complex_ridge_pipeline():
    """
    Return pipeline with the following structure:

    .. image:: img_ts_pipelines/ts_complex_ridge_pipeline.png
      :width: 55%

    """
    pip_builder = PipelineBuilder() \
        .add_sequence('lagged', 'ridge', branch_idx=0) \
        .add_sequence('lagged', 'ridge', branch_idx=1).join_branches('ridge')

    pipeline = pip_builder.to_pipeline()
    return pipeline


def ts_complex_ridge_smoothing_pipeline():
    """
    Pipeline looking like this

    .. image:: img_ts_pipelines/ts_complex_ridge_smoothing_pipeline.png
      :width: 55%

    Where smoothing - rolling mean
    """
    pip_builder = PipelineBuilder() \
        .add_sequence('smoothing', 'lagged', 'ridge', branch_idx=0) \
        .add_sequence('lagged', 'ridge', branch_idx=1).join_branches('ridge')

    pipeline = pip_builder.to_pipeline()
    return pipeline


def ts_complex_dtreg_pipeline(first_node='lagged'):
    """
    Return pipeline with the following structure:

    .. image:: img_ts_pipelines/ts_complex_dtreg_pipeline.png
      :width: 55%

    Where dtreg = tree regressor, rfr - random forest regressor
    """
    pip_builder = PipelineBuilder() \
        .add_sequence(first_node, 'dtreg', branch_idx=0) \
        .add_sequence(first_node, 'dtreg', branch_idx=1).join_branches('rfr')

    pipeline = pip_builder.to_pipeline()
    return pipeline


def ts_multiple_ets_pipeline():
    """
    Return pipeline with the following structure:

    .. image:: img_ts_pipelines/ts_multiple_ets_pipeline.png
      :width: 55%

    Where ets - exponential_smoothing
    """
    pip_builder = PipelineBuilder() \
        .add_sequence('ets', branch_idx=0) \
        .add_sequence('ets', branch_idx=1) \
        .add_sequence('ets', branch_idx=2) \
        .join_branches('lasso')

    pipeline = pip_builder.to_pipeline()
    return pipeline


def ts_ar_pipeline():
    """
    Return pipeline with the following structure:

    .. image:: img_ts_pipelines/ts_ar_pipeline.png
      :width: 55%

    Where ar - auto regression
    """
    pipeline = PipelineBuilder().add_node('ar').to_pipeline()
    return pipeline


def ts_arima_pipeline():
    """
    Return pipeline with the following structure:

    .. image:: img_ts_pipelines/ts_arima_pipeline.png
      :width: 55%

    """
    pipeline = PipelineBuilder().add_node("arima").to_pipeline()
    return pipeline


def ts_stl_arima_pipeline():
    """
    Return pipeline with the following structure:

    .. image:: img_ts_pipelines/ts_stl_arima_pipeline.png
      :width: 55%

    """
    pipeline = PipelineBuilder().add_node("stl_arima").to_pipeline()
    return pipeline


def ts_locf_ridge_pipeline():
    """
    Pipeline with naive LOCF (last observation carried forward) model
    and lagged features

    .. image:: img_ts_pipelines/ts_locf_ridge_pipeline.png
      :width: 55%

    """
    pip_builder = PipelineBuilder() \
        .add_sequence('locf', branch_idx=0) \
        .add_sequence('lagged', branch_idx=1) \
        .join_branches('ridge')

    pipeline = pip_builder.to_pipeline()
    return pipeline


def ts_naive_average_ridge_pipeline():
    """
    Pipeline with simple forecasting model (the forecast is mean value for known
    part)

    .. image:: img_ts_pipelines/ts_naive_average_ridge_pipeline.png
      :width: 55%

    """
    pip_builder = PipelineBuilder() \
        .add_sequence('ts_naive_average', branch_idx=0) \
        .add_sequence('lagged', branch_idx=1) \
        .join_branches('ridge')

    pipeline = pip_builder.to_pipeline()
    return pipeline


def clstm_pipeline():
    """
    Return pipeline with the following structure:

    .. image:: img_ts_pipelines/clstm_pipeline.png
      :width: 55%

    Where clstm - convolutional long short-term memory model
    """
    clstm_params = {'window_size': 29, 'hidden_size': 50, 'learning_rate': 0.004,
                    'cnn1_kernel_size': 5, 'cnn1_output_size': 32,
                    'cnn2_kernel_size': 4, 'cnn2_output_size': 32,
                    'batch_size': 64, 'num_epochs': 3, 'teacher_forcing': 0.8,
                    'optimizer': 'adam', 'loss': 'mse'}

    pip_builder = PipelineBuilder() \
        .add_sequence('lagged', 'ridge', branch_idx=0) \
        .add_sequence(('clstm', clstm_params), branch_idx=1) \
        .join_branches('ridge')

    pipeline = pip_builder.to_pipeline()
    return pipeline



from datetime import timedelta

from fedot.core.repository.operation_types_repository import OperationTypesRepository
from sklearn.metrics import roc_auc_score as roc_auc

from cases.data.data_utils import get_scoring_case_data_paths
from fedot.core.data.data import InputData
from fedot.core.pipelines.node import PrimaryNode, SecondaryNode
from fedot.core.pipelines.pipeline import Pipeline


# TODO not working now - add switch to other repository.json
def run_pipeline_from_automl(train_file_path: str, test_file_path: str,
                             max_run_time: timedelta = timedelta(minutes=10)):
    """ Function run pipeline with Auto ML models in nodes

    :param train_file_path: path to the csv file with data for train
    :param test_file_path: path to the csv file with data for validation
    :param max_run_time: maximum running time for customization of the "tpot" model

    :return roc_auc_value: ROC AUC metric for pipeline
    """
    with OperationTypesRepository.init_automl_repository() as _:
        train_data = InputData.from_csv(train_file_path)
        test_data = InputData.from_csv(test_file_path)

        testing_target = test_data.target

        node_scaling = PrimaryNode('scaling')
        node_tpot = PrimaryNode('tpot_class')

        node_tpot.parameters = {'timeout': max_run_time.seconds}

        node_lda = SecondaryNode('lda', nodes_from=[node_scaling])
        node_rf = SecondaryNode('rf', nodes_from=[node_tpot, node_lda])
        pipeline = Pipeline(node_rf)

        pipeline.fit(train_data)
        results = pipeline.predict(test_data)

        roc_auc_value = roc_auc(y_true=testing_target,
                                y_score=results.predict)
        print(roc_auc_value)
    return roc_auc_value


if __name__ == '__main__':
    train_file_path, test_file_path = get_scoring_case_data_paths()
    run_pipeline_from_automl(train_file_path, test_file_path)



import ast
import configparser
import os
from typing import Union

from fedot.core.repository.tasks import Task, TaskTypesEnum, TsForecastingParams
from fedot.core.utils import fedot_project_root

tmp_task = Task(TaskTypesEnum.ts_forecasting, TsForecastingParams(forecast_length=1))


class PipelineRunConfig:
    """
    Quasi-dataclass for the input parameters of external pipeline fitting
    """

    def __init__(self, config=None):
        if config is None:
            return

        self.pipeline_template = config['DEFAULT']['pipeline_template']
        self.input_data = config['DEFAULT']['train_data']
        self.task = eval(config['DEFAULT']['task'])
        self.output_path = config['DEFAULT']['output_path']

        self.train_data_idx = None
        if config['DEFAULT']['train_data_idx'] != 'None':
            self.train_data_idx = ast.literal_eval(config['DEFAULT']['train_data_idx'])

        self.is_multi_modal = False
        if config['DEFAULT']['is_multi_modal'] != 'None':
            self.is_multi_modal = ast.literal_eval(config['DEFAULT']['is_multi_modal'])

        self.var_names = False
        if config['DEFAULT']['var_names'] != 'None':
            self.var_names = ast.literal_eval(config['DEFAULT']['var_names'])

        self.target = None
        if 'target' in config['DEFAULT'] and config['DEFAULT']['target'] != 'None':
            try:
                # list of target values
                self.target = ast.literal_eval(config['DEFAULT']['target'])
            except ValueError:
                # name of target column
                self.target = config['DEFAULT']['target']

        self.test_data_path = config['OPTIONAL'].get('test_data')

    def load_from_file(self, file: Union[str, bytes]):
        config = configparser.ConfigParser()

        if isinstance(file, bytes):
            config.read_string(file.decode('utf-8'))
        else:
            if not os.path.exists(file):
                raise ValueError('Config not found')
            config.read(file, encoding='utf-8')

        processed_config = PipelineRunConfig(config)

        if '{fedot_base_path}' in processed_config.input_data:
            processed_config.input_data = \
                processed_config.input_data.format(fedot_base_path=fedot_project_root())

        return processed_config



import warnings
from typing import Optional

from fedot.core.data.data import InputData, OutputData
from fedot.core.operations.evaluation.evaluation_interfaces import EvaluationStrategy
from fedot.core.operations.evaluation.operation_implementations.data_operations.ts_transformations import \
    ExogDataTransformationImplementation, GaussianFilterImplementation, LaggedTransformationImplementation, \
    TsSmoothingImplementation, SparseLaggedTransformationImplementation, CutImplementation, \
    NumericalDerivativeFilterImplementation
from fedot.core.operations.evaluation.operation_implementations.models.ts_implementations.naive import \
    RepeatLastValueImplementation, NaiveAverageForecastImplementation
from fedot.core.operations.evaluation.operation_implementations.models. \
    ts_implementations.statsmodels import AutoRegImplementation, GLMImplementation, ExpSmoothingImplementation
from fedot.core.operations.evaluation.operation_implementations.models.ts_implementations.arima import \
    ARIMAImplementation, STLForecastARIMAImplementation
from fedot.core.operations.evaluation.operation_implementations.models.ts_implementations.clstm import \
    CLSTMImplementation
from fedot.core.operations.evaluation.operation_implementations.models.ts_implementations.poly import \
    PolyfitImplementation
from fedot.core.operations.operation_parameters import OperationParameters

warnings.filterwarnings("ignore", category=UserWarning)


class FedotTsForecastingStrategy(EvaluationStrategy):
    """
    This class defines the certain classical models implementation for time
    series forecasting (e.g. AR, ARIMA)

    :param str operation_type: str type of the operation defined in operation or
    data operation repositories
    :param dict params: hyperparameters to fit the model with
    """

    __operations_by_types = {
        'arima': ARIMAImplementation,
        'ar': AutoRegImplementation,
        'stl_arima': STLForecastARIMAImplementation,
        'ets': ExpSmoothingImplementation,
        'clstm': CLSTMImplementation,
        'polyfit': PolyfitImplementation,
        'glm': GLMImplementation,
        'locf': RepeatLastValueImplementation,
        'ts_naive_average': NaiveAverageForecastImplementation
    }

    def __init__(self, operation_type: str, params: Optional[OperationParameters] = None):
        super().__init__(operation_type, params)
        self.operation = self._convert_to_operation(operation_type)

    def fit(self, train_data: InputData):
        """
        This method is used for operation training with the data provided
        :param InputData train_data: data used for operation training
        :return: trained model
        """
        warnings.filterwarnings("ignore", category=RuntimeWarning)
        model = self.operation(self.params_for_fit)

        model.fit(train_data)
        return model

    def predict(self, trained_operation, predict_data: InputData) -> OutputData:
        """
        This method used for prediction of the target data during predict stage.

        :param trained_operation: trained operation object
        :param predict_data: data to predict
        :return OutputData: passed data with new predicted target
        """

        prediction = trained_operation.predict(predict_data)
        converted = self._convert_to_output(prediction, predict_data)
        return converted

    def predict_for_fit(self, trained_operation, predict_data: InputData) -> OutputData:
        """
        This method used for prediction of the target data during fit stage.

        :param trained_operation: trained operation object
        :param predict_data: data to predict
        :return OutputData: passed data with new predicted target
        """

        prediction = trained_operation.predict_for_fit(predict_data)
        converted = self._convert_to_output(prediction, predict_data)
        return converted

    def _convert_to_operation(self, operation_type: str):
        if operation_type in self.__operations_by_types.keys():
            return self.__operations_by_types[operation_type]
        else:
            raise ValueError(f'Impossible to obtain custom time series forecasting '
                             f'strategy for {operation_type}')


class FedotTsTransformingStrategy(EvaluationStrategy):
    """
    This class defines the certain data operation implementation for time series
    forecasting

    :param str operation_type: str type of the operation defined in operation or
    data operation repositories
    :param dict params: hyperparameters to fit the model with
    """

    __operations_by_types = {
        'lagged': LaggedTransformationImplementation,
        'sparse_lagged': SparseLaggedTransformationImplementation,
        'smoothing': TsSmoothingImplementation,
        'exog_ts': ExogDataTransformationImplementation,
        'gaussian_filter': GaussianFilterImplementation,
        'diff_filter': NumericalDerivativeFilterImplementation,
        'cut': CutImplementation}

    def __init__(self, operation_type: str, params: Optional[OperationParameters] = None):
        super().__init__(operation_type, params)
        self.operation = self._convert_to_operation(self.operation_type)

    def fit(self, train_data: InputData):
        """
        This method is used for operation training with the data provided
        :param InputData train_data: data used for operation training
        :return: trained operation (if it is needed for applying)
        """
        warnings.filterwarnings("ignore", category=RuntimeWarning)
        transformation_operation = self.operation(self.params_for_fit)
        transformation_operation.fit(train_data)
        return transformation_operation

    def predict(self, trained_operation, predict_data: InputData) -> OutputData:
        """
        This method used for prediction of the target data during predict stage.

        :param trained_operation: trained operation object
        :param predict_data: data to predict
        :return OutputData: passed data with new predicted target
        """

        prediction = trained_operation.transform(predict_data)
        converted = self._convert_to_output(prediction, predict_data)
        return converted

    def predict_for_fit(self, trained_operation, predict_data: InputData) -> OutputData:
        """
        This method used for prediction of the target data during fit stage.

        :param trained_operation: trained operation object
        :param predict_data: data to predict
        :return OutputData: passed data with new predicted target
        """

        prediction = trained_operation.transform_for_fit(predict_data)
        converted = self._convert_to_output(prediction, predict_data)
        return converted

    def _convert_to_operation(self, operation_type: str):
        if operation_type in self.__operations_by_types.keys():
            return self.__operations_by_types[operation_type]
        else:
            raise ValueError(f'Impossible to obtain custom time series transforming '
                             f'strategy for {operation_type}')



import sys
from abc import abstractmethod

import numpy as np
from sklearn.metrics import (accuracy_score, f1_score, log_loss, mean_absolute_error, mean_absolute_percentage_error,
                             mean_squared_error, mean_squared_log_error, precision_score, r2_score, roc_auc_score,
                             silhouette_score, roc_curve, auc)

from fedot.core.data.data import InputData, OutputData
from fedot.core.repository.dataset_types import DataTypesEnum
from fedot.core.repository.tasks import TaskTypesEnum
from fedot.core.pipelines.ts_wrappers import in_sample_ts_forecast
from fedot.core.pipelines.pipeline import Pipeline


def from_maximised_metric(metric_func):
    def wrapper(*args, **kwargs):
        return -metric_func(*args, **kwargs)

    return wrapper


def smape(y_true: np.ndarray, y_pred: np.ndarray) -> float:
    """ Symmetric mean absolute percentage error """

    numerator = 2 * np.abs(y_true - y_pred)
    denominator = np.abs(y_true) + np.abs(y_pred)
    result = numerator / denominator
    result[np.isnan(result)] = 0.0
    return float(np.mean(100 * result))


class Metric:
    output_mode = 'default'
    default_value = 0

    @classmethod
    @abstractmethod
    def get_value(cls, pipeline: 'Pipeline', reference_data: InputData,
                  validation_blocks: int) -> float:
        """ Get metrics values based on pipeline and InputData for validation """
        raise NotImplementedError()

    @staticmethod
    @abstractmethod
    def metric(reference: InputData, predicted: OutputData) -> float:
        raise NotImplementedError()


class QualityMetric:
    max_penalty_part = 0.01
    output_mode = 'default'
    default_value = 0

    @staticmethod
    @abstractmethod
    def metric(reference: InputData, predicted: OutputData) -> float:
        raise NotImplementedError()

    @classmethod
    def get_value(cls, pipeline: 'Pipeline', reference_data: InputData,
                  validation_blocks: int = None) -> float:
        metric = cls.default_value
        try:
            if validation_blocks is None:
                # Time series or regression classical hold-out validation
                results, reference_data = cls._simple_prediction(pipeline, reference_data)
            else:
                # Perform time series in-sample validation
                reference_data, results = cls._in_sample_prediction(pipeline, reference_data, validation_blocks)
            metric = cls.metric(reference_data, results)
        except Exception as ex:
            # TODO: use log instead of stdout
            print(f'Metric evaluation error: {ex}')

        return metric

    @classmethod
    def _simple_prediction(cls, pipeline: 'Pipeline', reference_data: InputData):
        """ Method prepares data for metric evaluation and perform simple validation """
        results = pipeline.predict(reference_data, output_mode=cls.output_mode)

        # Define conditions for target and predictions transforming
        is_regression = reference_data.task.task_type == TaskTypesEnum.regression
        is_multi_target = len(np.array(results.predict).shape) > 1
        is_multi_target_regression = is_regression and is_multi_target

        # Time series forecasting
        is_ts_forecasting = reference_data.task.task_type == TaskTypesEnum.ts_forecasting
        if is_ts_forecasting or is_multi_target_regression:
            results, reference_data = cls.flatten_convert(results, reference_data)

        return results, reference_data

    @staticmethod
    def flatten_convert(results, reference_data):
        """ Transform target and predictions by converting them into
        one-dimensional array

        :param results: output from pipeline
        :param reference_data: actual data for validation
        """
        # Predictions convert into uni-variate array
        forecast_values = np.ravel(np.array(results.predict))
        results.predict = forecast_values
        # Target convert into uni-variate array
        target_values = np.ravel(np.array(reference_data.target))
        reference_data.target = target_values

        return results, reference_data

    @classmethod
    def get_value_with_penalty(cls, pipeline: 'Pipeline', reference_data: InputData,
                               validation_blocks: int = None) -> float:
        quality_metric = cls.get_value(pipeline, reference_data)
        structural_metric = StructuralComplexity.get_value(pipeline)

        penalty = abs(structural_metric * quality_metric * cls.max_penalty_part)
        metric_with_penalty = (quality_metric +
                               min(penalty, abs(quality_metric * cls.max_penalty_part)))
        return metric_with_penalty

    @staticmethod
    def _in_sample_prediction(pipeline, data, validation_blocks):
        """ Performs in-sample pipeline validation for time series prediction """

        horizon = int(validation_blocks * data.task.task_params.forecast_length)

        actual_values = data.target[-horizon:]

        predicted_values = in_sample_ts_forecast(pipeline=pipeline,
                                                 input_data=data,
                                                 horizon=horizon)

        # Wrap target and prediction arrays into OutputData and InputData
        results = OutputData(idx=np.arange(0, len(predicted_values)), features=predicted_values,
                             predict=predicted_values, task=data.task, target=predicted_values,
                             data_type=DataTypesEnum.ts)
        reference_data = InputData(idx=np.arange(0, len(actual_values)), features=actual_values,
                                   task=data.task, target=actual_values, data_type=DataTypesEnum.ts)

        return reference_data, results


class RMSE(QualityMetric):
    default_value = sys.maxsize

    @staticmethod
    def metric(reference: InputData, predicted: OutputData) -> float:
        return mean_squared_error(y_true=reference.target,
                                  y_pred=predicted.predict, squared=False)


class MSE(QualityMetric):
    default_value = sys.maxsize

    @staticmethod
    def metric(reference: InputData, predicted: OutputData) -> float:
        return mean_squared_error(y_true=reference.target,
                                  y_pred=predicted.predict, squared=True)


class MSLE(QualityMetric):
    default_value = sys.maxsize

    @staticmethod
    def metric(reference: InputData, predicted: OutputData) -> float:
        return mean_squared_log_error(y_true=reference.target,
                                      y_pred=predicted.predict)


class MAPE(QualityMetric):
    default_value = sys.maxsize

    @staticmethod
    def metric(reference: InputData, predicted: OutputData) -> float:
        return mean_absolute_percentage_error(y_true=reference.target,
                                              y_pred=predicted.predict)


class SMAPE(QualityMetric):
    default_value = sys.maxsize

    @staticmethod
    def metric(reference: InputData, predicted: OutputData) -> float:
        return smape(y_true=reference.target, y_pred=predicted.predict)


class F1(QualityMetric):
    default_value = 0
    output_mode = 'labels'

    @staticmethod
    @from_maximised_metric
    def metric(reference: InputData, predicted: OutputData) -> float:
        n_classes = reference.num_classes
        if n_classes > 2:
            additional_params = {'average': 'weighted'}
        else:
            additional_params = {'average': 'micro'}
        return f1_score(y_true=reference.target, y_pred=predicted.predict,
                        **additional_params)


class MAE(QualityMetric):
    default_value = sys.maxsize

    @staticmethod
    def metric(reference: InputData, predicted: OutputData) -> float:
        return mean_absolute_error(y_true=reference.target, y_pred=predicted.predict)


class R2(QualityMetric):
    default_value = 0

    @staticmethod
    def metric(reference: InputData, predicted: OutputData) -> float:
        return r2_score(y_true=reference.target, y_pred=predicted.predict)


class ROCAUC(QualityMetric):
    default_value = 0.5

    @staticmethod
    @from_maximised_metric
    def metric(reference: InputData, predicted: OutputData) -> float:
        n_classes = reference.num_classes
        if n_classes > 2:
            additional_params = {'multi_class': 'ovr', 'average': 'macro'}
        else:
            additional_params = {}

        score = round(roc_auc_score(y_score=predicted.predict,
                                    y_true=reference.target,
                                    **additional_params), 3)

        return score

    @staticmethod
    def roc_curve(target: np.ndarray, predict: np.ndarray, pos_label=None):

        return roc_curve(target, predict, pos_label=pos_label)

    @classmethod
    def auc(cls, fpr, tpr):
        return auc(fpr, tpr)


class Precision(QualityMetric):
    output_mode = 'labels'

    @staticmethod
    @from_maximised_metric
    def metric(reference: InputData, predicted: OutputData) -> float:
        return precision_score(y_true=reference.target, y_pred=predicted.predict)


class Logloss(QualityMetric):
    default_value = sys.maxsize

    @staticmethod
    def metric(reference: InputData, predicted: OutputData) -> float:
        return log_loss(y_true=reference.target, y_pred=predicted.predict)


class Accuracy(QualityMetric):
    default_value = 0
    output_mode = 'labels'

    @staticmethod
    @from_maximised_metric
    def metric(reference: InputData, predicted: OutputData) -> float:
        return accuracy_score(y_true=reference.target, y_pred=predicted.predict)


class Silhouette(QualityMetric):
    default_value = 1

    @staticmethod
    @from_maximised_metric
    def metric(reference: InputData, predicted: OutputData) -> float:
        return silhouette_score(reference.features, labels=predicted.predict)


class StructuralComplexity(Metric):
    @classmethod
    def get_value(cls, pipeline: 'Pipeline', **args) -> float:
        norm_constant = 30
        return (pipeline.depth ** 2 + pipeline.length) / norm_constant


class NodeNum(Metric):
    @classmethod
    def get_value(cls, pipeline: 'Pipeline', **args) -> float:
        norm_constant = 10
        return pipeline.length / norm_constant


class ComputationTime(Metric):
    @classmethod
    def get_value(cls, pipeline: 'Pipeline', **args) -> float:
        return pipeline.computation_time



from copy import deepcopy
from typing import Optional, Iterable

from fedot.core.repository.default_params_repository import DefaultOperationParamsRepository


class OperationParameters:
    """Stores parameters for models and data operations implementations and records what parameters were changed.
    Uses operation_type to set default parameters from default_parameters_repository if a parameter was not passed
    with parameters.
    If neither parameters nor operation_type were passed, parameters are equal to empty dict.

    Args:
        operation_type: type of the operation defined in operation repository
            the custom prefix can be added after ``/`` (to highlight the specific node)\n
            **The prefix will be ignored at Implementation stage**
        parameters: dict with parameters

    """
    def __init__(self, **parameters):
        self._parameters = parameters
        self._changed_keys: list = []

    def __bool__(self):
        return bool(self._parameters)

    @staticmethod
    def from_operation_type(operation_type: str, **parameters):
        default_parameters = get_default_params(operation_type)
        parameters = {**default_parameters, **parameters}
        return OperationParameters(**parameters)

    def update(self, **params):
        for key, value in params.items():
            if key not in self._changed_keys:
                if self._parameters.get(key) != value:
                    self._changed_keys.append(key)
            self._parameters.update({key: value})

    def get(self, key, default_value=None):
        return self._parameters.get(key, default_value)

    def setdefault(self, key, value):
        if key not in self._parameters.keys():
            self.update(**{key: value})
        return self.get(key)

    def to_dict(self) -> dict:
        return deepcopy(self._parameters)

    def keys(self) -> Iterable:
        return self._parameters.keys()

    @property
    def changed_parameters(self) -> dict:
        changed_parameters = {key: self._parameters[key] for key in self._changed_keys}
        return changed_parameters


def get_default_params(operation_type: str) -> dict:
    """Gets default params for chosen model name

    :param operation_type: the operation name to choose default parameters for

    :return: default repository parameters for the model name
    """
    with DefaultOperationParamsRepository() as default_params_repo:
        return default_params_repo.get_default_params_for_operation(operation_type)



import os
from copy import deepcopy
from inspect import signature
from typing import Optional

from matplotlib import pyplot as plt
from sklearn import tree
from sklearn.tree._tree import TREE_LEAF

from fedot.core.composer.metrics import Metric
from fedot.core.composer.metrics import R2, F1
from fedot.core.data.data import InputData
from fedot.core.pipelines.node import PrimaryNode
from fedot.core.pipelines.pipeline import Pipeline
from fedot.core.repository.tasks import TaskTypesEnum
from fedot.explainability.explainer_template import Explainer


class SurrogateExplainer(Explainer):
    """
    Base class used for composite model structure definition

    :param model: `Pipeline` object to be explained
    :param surrogate: surrogate name. Supported surrogates: `[dt, dtreg]`

    .. note::
        `score` stores the score of surrogate's prediction on model (equals None if the 'explain' method hasn't been
        called yet)
    """

    surrogates_default_params = {
        'dt': {'max_depth': 5},
        'dtreg': {'max_depth': 5},
    }

    def __init__(self, model: 'Pipeline', surrogate: str):
        super().__init__(model)

        self.score: Optional[float] = None

        if not isinstance(surrogate, str):
            raise ValueError(f'{surrogate} is not supported as a surrogate model')
        if surrogate not in self.surrogates_default_params:
            raise ValueError(f'{type(surrogate)} is not supported as a surrogate model')

        self.surrogate_str = surrogate
        self.surrogate = get_simple_pipeline(self.surrogate_str, self.surrogates_default_params[surrogate])

    def explain(self, data: InputData, visualization: bool = False, **kwargs):
        try:
            self.score = fit_naive_surrogate_model(self.model, self.surrogate, data)

        except Exception as ex:
            print(f'Failed to fit the surrogate: {ex}')
            return

        # Pruning redundant branches and leaves
        if self.surrogate_str in ('dt', 'dtreg'):
            prune_duplicate_leaves(self.surrogate.root_node.fitted_operation)

        if visualization:
            self.visualize(**kwargs)

    def visualize(self, dpi: int = 100, figsize=(48, 12), save_path: str = None, **kwargs):
        """Print and plot results of the last explanation. Suitable keyword parameters
        are passed to the corresponding plot function.
        :param dpi: the figure DPI, defaults to 100.
        :param figsize: the figure size in format `(width, height)`, defaults to `(48, 12)`.
        :param save_path: path to save the plot.
        """
        plt.figure(dpi=dpi, figsize=figsize)
        if self.surrogate_str in ['dt', 'dtreg']:

            if self.score is not None:
                print(f'Surrogate\'s model reproduction quality: {self.score}')
            # Plot default parameters
            plot_params = {
                'proportion': True,
                'filled': True,
                'rounded': True,
                'fontsize': 12,
            }
            # Plot parameters defined by user
            kwargs_params = \
                {par: kwargs[par] for par in kwargs if par in signature(tree.plot_tree).parameters}

            plot_params.update(kwargs_params)

            tree.plot_tree(self.surrogate.root_node.fitted_operation, **plot_params)

        if save_path is not None:
            plt.savefig(save_path)
            print(f'Saved the plot to "{os.path.abspath(save_path)}"')


def get_simple_pipeline(model: str, custom_params: dict = None) -> 'Pipeline':
    surrogate_node = PrimaryNode(model)
    if custom_params:
        surrogate_node.parameters = custom_params
    return Pipeline(surrogate_node)


def fit_naive_surrogate_model(
        black_box_model: 'Pipeline', surrogate_model: 'Pipeline', data: 'InputData',
        metric: 'Metric' = None) -> Optional[float]:
    output_mode = 'default'

    if data.task.task_type == TaskTypesEnum.classification:
        output_mode = 'labels'
        if metric is None:
            metric = F1
    elif data.task.task_type == TaskTypesEnum.regression and metric is None:
        metric = R2

    prediction = black_box_model.predict(data, output_mode=output_mode)
    data.target = prediction.predict
    surrogate_model.fit(data)

    data_c = deepcopy(data)
    data_c.target = surrogate_model.predict(data, output_mode=output_mode).predict
    score = round(abs(metric.metric(data_c, prediction)), 2)

    return score


def is_leaf(inner_tree, index):
    # Check whether node is leaf node
    return (inner_tree.children_left[index] == TREE_LEAF and
            inner_tree.children_right[index] == TREE_LEAF)


def prune_index(inner_tree, decisions, index=0):
    # Start pruning from the bottom - if we start from the top, we might miss
    # nodes that become leaves during pruning.
    # Do not use this directly - use prune_duplicate_leaves instead.
    if not is_leaf(inner_tree, inner_tree.children_left[index]):
        prune_index(inner_tree, decisions, inner_tree.children_left[index])
    if not is_leaf(inner_tree, inner_tree.children_right[index]):
        prune_index(inner_tree, decisions, inner_tree.children_right[index])

    # Prune children if both children are leaves now and make the same decision:
    if (is_leaf(inner_tree, inner_tree.children_left[index]) and
            is_leaf(inner_tree, inner_tree.children_right[index]) and
            (decisions[index] == decisions[inner_tree.children_left[index]]) and
            (decisions[index] == decisions[inner_tree.children_right[index]])):
        # turn node into a leaf by "unlinking" its children
        inner_tree.children_left[index] = TREE_LEAF
        inner_tree.children_right[index] = TREE_LEAF


def prune_duplicate_leaves(mdl):
    """
    Function for pruning redundant leaves of a tree by Thomas (https://stackoverflow.com/users/4629950/thomas).
    Source: https://stackoverflow.com/questions/51397109/prune-unnecessary-leaves-in-sklearn-decisiontreeclassifier
    :param mdl: `DecisionTree` or `DecisionTreeRegressor` instance by sklearn.
    """
    # Remove leaves if both
    decisions = mdl.tree_.value.argmax(axis=2).flatten().tolist()  # Decision for each node
    prune_index(mdl.tree_, decisions)



import warnings
from typing import Optional

from fedot.core.utilities.random import RandomStateHandler
from fedot.utilities.requirements_notificator import warn_requirement

try:
    import cudf
    import cuml
    from cuml import Ridge, LogisticRegression, Lasso, ElasticNet, \
        MBSGDClassifier, MBSGDRegressor, CD
    from cuml.ensemble import RandomForestClassifier, RandomForestRegressor
    from cuml.svm import SVC
    from cuml.neighbors import KNeighborsClassifier as CuMlknnClassifier, \
        KNeighborsRegressor as CuMlknnRegressor
    from cuml import LinearRegression as CuMlLinReg, SGD as CuMlSGD, \
        MultinomialNB as CuMlMultinomialNB
except ModuleNotFoundError:
    warn_requirement('cudf / cuml')
    cudf = None
    cuml = None

from fedot.core.data.data import InputData, OutputData
from fedot.core.operations.evaluation.evaluation_interfaces import SkLearnEvaluationStrategy
from fedot.core.repository.operation_types_repository import OperationTypesRepository
from fedot.core.repository.tasks import TaskTypesEnum


class CuMLEvaluationStrategy(SkLearnEvaluationStrategy):
    """
    This class defines the certain operation implementation for the GPU-based CuML operations
    defined in operation repository
    :param str operation_type: str type of the operation defined in operation or
    data operation repositories
    :param dict params: hyperparameters to fit the operation with
    """
    try:
        __operations_by_types = {
            'ridge': Ridge,
            'lasso': Lasso,
            'logit': LogisticRegression,
            'linear': CuMlLinReg,
            'rf': RandomForestClassifier,
            'rfr': RandomForestRegressor,
            'svc': SVC,
            'knn': CuMlknnClassifier,
            'knnreg': CuMlknnRegressor,
            'sgd': CuMlSGD,
            'multinb': CuMlMultinomialNB,
            'elasticnet': ElasticNet,
            'mbsgdclass': MBSGDClassifier,
            'mbsgdcregr': MBSGDRegressor,
            'cd': CD
        }
    except NameError:
        # if cuML not installed
        __operations_by_types = {}

    def __init__(self, operation_type: str, params: Optional[dict] = None):
        super().__init__(operation_type, params)
        self.operation_impl = self._convert_to_operation(operation_type)
        cuml.set_global_output_type('numpy')

    def fit(self, train_data: InputData):
        """
        This method is used for operation training with the data provided
        :param InputData train_data: data used for operation training
        :return: trained cuML operation
        """
        warnings.filterwarnings("ignore", category=RuntimeWarning)
        operation_implementation = self.operation_impl(**self.params_for_fit.to_dict())

        # If model doesn't support multi-output and current task is ts_forecasting
        current_task = train_data.task.task_type
        models_repo = OperationTypesRepository()
        non_multi_models = models_repo.suitable_operation(task_type=current_task,
                                                          tags=['non_multi'])
        is_model_not_support_multi = self.operation_type in non_multi_models
        features = cudf.DataFrame(train_data.features.astype('float32'))
        target = cudf.Series(train_data.target.flatten().astype('float32'))

        if is_model_not_support_multi and current_task == TaskTypesEnum.ts_forecasting:
            raise NotImplementedError('Not supported for GPU yet')
            # TODO Manually wrap the regressor into multi-output model
        else:
            with RandomStateHandler():
                operation_implementation.fit(features, target)
        return operation_implementation

    def predict(self, trained_operation, predict_data: InputData) -> OutputData:
        """
        This method used for prediction of the target data during predict stage.
        :param trained_operation: operation object
        :param predict_data: data to predict
        :return OutputData: passed data with new predicted target
        """
        raise NotImplementedError()

    def _convert_to_operation(self, operation_type: str):
        if operation_type in self.__operations_by_types.keys():
            return self.__operations_by_types[operation_type]
        else:
            raise ValueError(f'Impossible to obtain cuML strategy for {operation_type}')



from examples.advanced.sensitivity_analysis.dataset_access import get_scoring_data, get_kc2_data, get_cholesterol_data
from fedot.core.repository.quality_metrics_repository import MetricsRepository, ClassificationMetricsEnum, \
    RegressionMetricsEnum
from fedot.core.repository.tasks import Task, TaskTypesEnum
from examples.advanced.sensitivity_analysis.case_analysis import run_case_analysis


def run_class_scoring_case(sa_class: str, is_composed: bool = False, path_to_save=None):
    train_data, test_data = get_scoring_data()
    task = Task(TaskTypesEnum.classification)
    # the choice of the metric for the pipeline quality assessment during composition
    metric_function = MetricsRepository().metric_by_id(ClassificationMetricsEnum.ROCAUC_penalty)

    if is_composed:
        case = f'scoring_composed_{sa_class}'
        is_composed = True
    else:
        case = f'scoring_{sa_class}'
        is_composed = False

    run_case_analysis(train_data=train_data,
                      test_data=test_data,
                      case_name=case, task=task, sa_class=sa_class,
                      metric=metric_function,
                      is_composed=is_composed, result_path=path_to_save)


def run_class_kc2_case(sa_class: str, is_composed: bool = False, path_to_save=None):
    train_data, test_data = get_kc2_data()
    task = Task(TaskTypesEnum.classification)
    # the choice of the metric for the pipeline quality assessment during composition
    metric_function = MetricsRepository().metric_by_id(ClassificationMetricsEnum.ROCAUC_penalty)

    if is_composed:
        case = f'kc2_composed_{sa_class}'
        is_composed = True
    else:
        case = f'kc2_{sa_class}'
        is_composed = False

    run_case_analysis(train_data=train_data,
                      test_data=test_data,
                      case_name=case, task=task, sa_class=sa_class,
                      metric=metric_function,
                      is_composed=is_composed, result_path=path_to_save)


def run_regr_case(sa_class: str, is_composed: bool = False, path_to_save=None):
    train_data, test_data = get_cholesterol_data()
    task = Task(TaskTypesEnum.regression)
    # the choice of the metric for the pipeline quality assessment during composition
    metric_function = MetricsRepository().metric_by_id(RegressionMetricsEnum.RMSE)

    if is_composed:
        case = f'cholesterol_composed_{sa_class}'
        is_composed = True
    else:
        case = f'cholesterol_{sa_class}'
        is_composed = False

    run_case_analysis(train_data=train_data,
                      test_data=test_data,
                      case_name=case, task=task, sa_class=sa_class,
                      metric=metric_function,
                      is_composed=is_composed, result_path=path_to_save)


if __name__ == '__main__':
    # You can assign any of PipelineSensitivityAnalysis, NodesAnalysis, PipelineAnalysis to any case

    # scoring case
    run_class_scoring_case(is_composed=False, sa_class='PipelineSensitivityAnalysis')

    # kc2 case
    run_class_kc2_case(is_composed=False, sa_class='NodesAnalysis')

    # cholesterol case
    run_regr_case(is_composed=False, sa_class='PipelineAnalysis')



import re
from typing import Optional

import numpy as np

from fedot.core.data.data import InputData, OutputData
from fedot.core.operations.operation_parameters import OperationParameters
from fedot.utilities.requirements_notificator import warn_requirement

try:
    import nltk
except ModuleNotFoundError:
    warn_requirement('nltk')
    nltk = None

from fedot.core.operations.evaluation.operation_implementations.implementation_interfaces import (
    DataOperationImplementation
)
from fedot.core.repository.dataset_types import DataTypesEnum


class TextCleanImplementation(DataOperationImplementation):
    """ Class for text cleaning (lemmatization and stemming) operation """

    def __init__(self, params: Optional[OperationParameters]):
        super().__init__(params)
        self.stemmer = nltk.stem.SnowballStemmer(language=self.language)
        self.lemmatizer = nltk.stem.WordNetLemmatizer()
        self._download_nltk_resources()

    @property
    def language(self) -> str:
        return self.params.setdefault('language', 'english')

    def fit(self, input_data: InputData):
        """ Class doesn't support fit operation

        :param input_data: data with features, target and ids to process
        """
        pass

    def transform(self, input_data: InputData) -> OutputData:
        """ Method for transformation of the text data for predict stage

        :param input_data: data with features, target and ids to process
        :return output_data: output data with transformed features table
        """

        clean_data = []
        for text in input_data.features:
            text = str(text).lower()
            words = self._word_vectorize(text)
            without_stop_words = self._remove_stop_words(words)
            words = self._lemmatization(without_stop_words)
            words = [word for word in words if word.isalpha()]
            new_text = ' '.join(words)
            new_text = self._clean_html_text(new_text)
            clean_data.append(new_text)
        clean_data = np.array(clean_data)

        output_data = self._convert_to_output(input_data,
                                              clean_data,
                                              data_type=DataTypesEnum.text)
        return output_data

    @staticmethod
    def _download_nltk_resources():
        for resource in ['punkt']:
            try:
                nltk.data.find(f'tokenizers/{resource}')
            except LookupError:
                nltk.download(f'{resource}')
        for resource in ['stopwords', 'wordnet', 'omw-1.4']:
            try:
                nltk.data.find(f'corpora/{resource}')
            except LookupError:
                nltk.download(f'{resource}')

    @staticmethod
    def _word_vectorize(text):
        if isinstance(text, np.ndarray):
            # occurrs when node with text preprocessing is not primary
            text = text[0]
        words = nltk.word_tokenize(text)

        return words

    def _remove_stop_words(self, words: set):
        stop_words = set(nltk.corpus.stopwords.words(self.language))
        cleared_words = [word for word in words if word not in stop_words]

        return cleared_words

    def _stemming(self, words):
        stemmed_words = [self.stemmer.stem(word) for word in words]

        return stemmed_words

    def _lemmatization(self, words):
        # TODO pos
        lemmas = [self.lemmatizer.lemmatize(word) for word in words]

        return lemmas

    @staticmethod
    def _clean_html_text(raw_text):
        clean_pattern = re.compile('<.*?>')
        text = re.sub(clean_pattern, ' ', raw_text)

        return text



from fedot.utilities.requirements_notificator import warn_requirement

try:
    import cudf
except ModuleNotFoundError:
    warn_requirement('cudf')
    cudf = None

from fedot.core.data.data import InputData, OutputData
from fedot.core.operations.evaluation.gpu.common import CuMLEvaluationStrategy


class CuMLRegressionStrategy(CuMLEvaluationStrategy):
    def predict(self, trained_operation, predict_data: InputData) -> OutputData:
        """
        Predict method for regression task for predict stage
        :param trained_operation: model object
        :param predict_data: data used for prediction
        :return:
        """

        features = cudf.DataFrame(predict_data.features.astype('float32'))

        prediction = trained_operation.predict(features)
        converted = self._convert_to_output(prediction, predict_data)

        return converted



from copy import copy
from typing import Optional, Dict, Any

import numpy as np
import sklearn
from pkg_resources import parse_version
from sklearn.ensemble import IsolationForest
from sklearn.linear_model import LinearRegression, RANSACRegressor
from sklearn.pipeline import make_pipeline
from sklearn.preprocessing import StandardScaler
from sklearn.tree import DecisionTreeRegressor

from fedot.core.data.data import InputData, OutputData
from fedot.core.log import default_log
from fedot.core.operations.evaluation.operation_implementations.implementation_interfaces import (
    DataOperationImplementation
)
from fedot.core.operations.operation_parameters import OperationParameters


class FilterImplementation(DataOperationImplementation):
    """ Base class for applying filtering operations on tabular data """

    def __init__(self, params: Optional[OperationParameters]):
        super().__init__(params)
        self.inner_model = None
        self.operation = None

        self.log = default_log(self)

    def fit(self, input_data: InputData):
        """ Method for fit filter

        :param input_data: data with features, target and ids to process
        :return operation: trained operation (optional output)
        """

        self.operation.fit(input_data.features, input_data.target)

        return self.operation

    def transform(self, input_data: InputData) -> OutputData:
        """ Method for making prediction for predict stage

        :param input_data: data with features, target and ids to process
        :return output_data: filtered input data by rows
        """
        output_data = self._convert_to_output(input_data,
                                              input_data.features)
        return output_data

    def transform_for_fit(self, input_data: InputData) -> OutputData:
        """ Method for making prediction for fit stage

        :param input_data: data with features, target and ids to process
        :return output_data: filtered input data by rows
        """
        # For fit stage - filter data
        mask = self.operation.inlier_mask_
        if mask is not None:
            input_data = update_data(input_data, mask)
        else:
            self.log.info("Filtering Algorithm: didn't fit correctly. Return all objects")
        output_data = self._convert_to_output(input_data,
                                              input_data.features)
        return output_data


class RegRANSACImplementation(FilterImplementation):
    def __init__(self, params: Optional[OperationParameters]):
        super().__init__(params)
        self.max_iter = 10

    def fit(self, input_data: InputData):
        iter_ = 0

        while iter_ < self.max_iter:
            try:
                self.operation.inlier_mask_ = None
                self.operation.fit(input_data.features, input_data.target)
                return self.operation
            except ValueError:
                self.log.info("RANSAC: multiplied residual_threshold on 2")
                residual_threshold = self.params.get('residual_threshold')
                self.params.update(residual_threshold=residual_threshold * 2)
                iter_ += 1

        return self.operation


class LinearRegRANSACImplementation(RegRANSACImplementation):
    """
    RANdom SAmple Consensus (RANSAC) algorithm with LinearRegression as core
    Task type - regression
    """

    def __init__(self, params: Optional[OperationParameters]):
        super().__init__(params)
        self.inner_model = make_pipeline(StandardScaler(with_mean=False), LinearRegression())

        # TODO valer1435 | Delete this after removing compatibility with sklearn<1.1
        if parse_version(sklearn.__version__) < parse_version('1.1.0'):
            self.operation = RANSACRegressor(base_estimator=self.inner_model, **self.params.to_dict())
        else:
            self.operation = RANSACRegressor(estimator=self.inner_model, **self.params.to_dict())


class NonLinearRegRANSACImplementation(RegRANSACImplementation):
    """
    RANdom SAmple Consensus (RANSAC) algorithm with DecisionTreeRegressor as core
    Task type - regression
    """

    def __init__(self, params: Optional[OperationParameters]):
        super().__init__(params)
        self.inner_model = DecisionTreeRegressor()

        # TODO valer1435 | Delete this after removing compatibility with sklearn<1.1
        if parse_version(sklearn.__version__) < parse_version('1.1.0'):
            self.operation = RANSACRegressor(base_estimator=self.inner_model, **self.params.to_dict())
        else:
            self.operation = RANSACRegressor(estimator=self.inner_model, **self.params.to_dict())


class IsolationForestRegImplementation(DataOperationImplementation):
    """
    Isolation Forest algorithm based on ExtraTreeRegressor
    Task type - regression
    """

    def __init__(self, params: Optional[OperationParameters]):
        super().__init__(params)
        self.log = default_log(self)
        self.operation = IsolationForest(**self.params.to_dict())

    def fit(self, input_data: InputData) -> 'IsolationForest':
        """ Method for fit filter

        :param input_data: data with features, target and ids to process
        :return operation: trained operation (optional output)
        """

        self.operation.fit(input_data.features, input_data.target)

        return self.operation

    def _get_inlier_mask(self, input_data: InputData) -> np.ndarray:
        """ Method for making boolean mask of inliers classified as False

        :param input_data: data with features, target and ids to process
        :return mask: boolean mask of inliers classified as False
        """

        predictions = self.operation.predict(input_data.features)
        mask = predictions == 1
        return mask

    def transform(self, input_data: InputData) -> OutputData:
        """ Method for making prediction for predict stage

        :param input_data: data with features, target and ids to process
        :return output_data: filtered input data by rows
        """
        output_data = self._convert_to_output(input_data,
                                              input_data.features)
        return output_data

    def transform_for_fit(self, input_data: InputData) -> OutputData:
        """ Method for making prediction for fit stage

        :param input_data: data with features, target and ids to process
        :return output_data: filtered input data by rows
        """
        # For fit stage - filter data
        mask = self._get_inlier_mask(input_data)
        input_data = update_data(input_data, mask)

        output_data = self._convert_to_output(input_data,
                                              input_data.features)
        return output_data


class IsolationForestClassImplementation(IsolationForestRegImplementation):
    """
    Isolation Forest algorithm based on ExtraTreeRegressor
    Task type - classification
    """

    @staticmethod
    def _is_inlier_mask_correct(input_targets: np.ndarray, modified_targets: np.ndarray) -> bool:
        """ Method for checking if inlier mask is correct
        Inlier mask considered correct if after its application no class is completely or partially
        removed from the dataset by more than 80 percent

        :param input_targets: targets from input_data
        :param modified_targets: targets from modified_targets
        :return True if mask is correct otherwise False
        """

        if np.unique(input_targets).shape != np.unique(modified_targets).shape:
            return False
        for counts in zip(np.unique(input_targets, return_counts=True)[1],
                          np.unique(modified_targets, return_counts=True)[1]):
            if counts[0] > 5 * counts[1]:
                return False
        return True

    def transform_for_fit(self, input_data: InputData) -> OutputData:
        """ Method for making prediction for fit stage

        :param input_data: data with features, target and ids to process
        :return output_data: filtered input data by rows
        """
        # For fit stage - filter data
        mask = self._get_inlier_mask(input_data)
        modified_input_data = update_data(input_data, mask)
        if self._is_inlier_mask_correct(input_data.target, modified_input_data.target):
            input_data = modified_input_data

        output_data = self._convert_to_output(input_data,
                                              input_data.features)
        return output_data


def update_data(input_data: InputData, mask: np.ndarray) -> InputData:
    """ Method for updating target and features"""

    modified_input_data = copy(input_data)
    old_features = modified_input_data.features
    old_target = modified_input_data.target
    old_idx = modified_input_data.idx

    modified_input_data.features = old_features[mask]
    modified_input_data.target = old_target[mask]
    modified_input_data.idx = np.array(old_idx)[mask]

    return modified_input_data



import pandas as pd

from fedot.api.main import Fedot
from fedot.core.utils import fedot_project_root


def run_classification_multiobj_example(visualization=False, timeout=1):
    train_data = pd.read_csv(f'{fedot_project_root()}/examples/data/Hill_Valley_with_noise_Training.data')
    test_data = pd.read_csv(f'{fedot_project_root()}/examples/data/Hill_Valley_with_noise_Testing.data')
    target = test_data['class']
    del test_data['class']
    problem = 'classification'

    metric_names = ['f1', 'node_num']
    auto_model = Fedot(problem=problem, timeout=timeout, preset='best_quality', seed=42,
                       metric=metric_names,
                       with_tuning=False)
    auto_model.fit(features=train_data, target='class')
    prediction = auto_model.predict_proba(features=test_data)
    print(auto_model.get_metrics(target))

    if visualization:
        auto_model.plot_prediction()
        auto_model.plot_pareto()

    return prediction


if __name__ == '__main__':
    run_classification_multiobj_example(visualization=True)



import warnings

from fedot.core.operations.operation_parameters import OperationParameters
from fedot.core.utilities.random import RandomStateHandler
from fedot.utilities.requirements_notificator import warn_requirement

try:
    from cuml import KMeans
    import cudf
except ModuleNotFoundError:
    warn_requirement('cudf / cuml')
    cudf = None
    KMeans = None

from typing import Optional

from fedot.core.data.data import InputData, OutputData

from fedot.core.operations.evaluation.gpu.common import CuMLEvaluationStrategy


class CumlClusteringStrategy(CuMLEvaluationStrategy):
    __operations_by_types = {
        'kmeans': KMeans
    }

    def __init__(self, operation_type: str, params: Optional[OperationParameters] = None):
        super().__init__(operation_type, params)
        self.operation_impl = self._convert_to_operation(operation_type)

    def fit(self, train_data: InputData):
        """
        Fit method for clustering task

        :param train_data: data used for model training
        :return:
        """

        warnings.filterwarnings("ignore", category=RuntimeWarning)
        if self.params_for_fit:
            operation_implementation = self.operation_impl(**self.params_for_fit.to_dict())
        else:
            operation_implementation = self.operation_impl(n_clusters=2)

        features = cudf.DataFrame(train_data.features.astype('float32'))
        with RandomStateHandler():
            operation_implementation.fit(features)
        return operation_implementation

    def predict(self, trained_operation, predict_data: InputData) -> OutputData:
        """
        Predict method for regression task for predict stage
        :param trained_operation: model object
        :param predict_data: data used for prediction
        :return:
        """

        features = cudf.DataFrame(predict_data.features.astype('float32'))

        prediction = trained_operation.predict(features)
        converted = self._convert_to_output(prediction, predict_data)

        return converted

    def _convert_to_operation(self, operation_type: str):
        if operation_type in self.__operations_by_types.keys():
            return self.__operations_by_types[operation_type]
        else:
            raise ValueError(f'Impossible to obtain SkLearn clustering strategy for {operation_type}')



import json
import random
from abc import ABC, abstractmethod
from copy import deepcopy
from os import makedirs
from os.path import exists, join
from typing import List, Optional, Type, Union, Callable

import matplotlib.pyplot as plt

from fedot.core.data.data import InputData
from fedot.core.log import default_log
from fedot.core.pipelines.node import Node
from fedot.core.pipelines.pipeline import Pipeline
from fedot.core.pipelines.verification import verify_pipeline
from fedot.core.repository.operation_types_repository import OperationTypesRepository
from fedot.core.repository.quality_metrics_repository import MetricsRepository
from fedot.core.utils import default_fedot_data_dir
from fedot.sensitivity.sa_requirements import ReplacementAnalysisMetaParams, SensitivityAnalysisRequirements
from fedot.utilities.define_metric_by_task import MetricByTask


class NodeAnalysis:
    """
    Args:
        approaches: methods applied to nodes to modify the pipeline or analyze certain operations.
            Default: [``NodeDeletionAnalyze``, ``NodeTuneAnalyze``, ``NodeReplaceOperationAnalyze``]
        path_to_save: path to save results to. Default: ``~home/Fedot/sensitivity``
    """

    def __init__(self, approaches: Optional[List[Type['NodeAnalyzeApproach']]] = None,
                 approaches_requirements: SensitivityAnalysisRequirements = None,
                 path_to_save=None):

        self.approaches = [NodeDeletionAnalyze, NodeReplaceOperationAnalyze] if approaches is None else approaches

        self.path_to_save = \
            join(default_fedot_data_dir(), 'sensitivity', 'nodes_sensitivity') if path_to_save is None else path_to_save
        self.log = default_log(self)

        self.approaches_requirements = \
            SensitivityAnalysisRequirements() if approaches_requirements is None else approaches_requirements

    def analyze(self, pipeline: Pipeline, node: Node,
                train_data: InputData, test_data: InputData,
                is_save: bool = False) -> dict:

        """Method runs Node analysis within defined approaches

        Args:
            is_save: whether the certain node analysis result is needed to ba saved
            pipeline: :obj:`Pipeline` containing the analyzed :obj:`Node`
            node: :obj:`Node` object to analyze in :obj:`Pipeline`
            train_data: data used for :obj:`Pipeline` training
            test_data: data used for :obj:`Pipeline` validation

        Returns:
            dict:  :obj:`Node` analysis result per approach
        """

        results = dict()
        for approach in self.approaches:
            results[f'{approach.__name__}'] = \
                approach(pipeline=pipeline,
                         train_data=train_data,
                         test_data=test_data,
                         requirements=self.approaches_requirements,
                         path_to_save=self.path_to_save).analyze(node=node)

        # TODO remove conflict with requirements.is_save
        if is_save:
            self._save_results_to_json(node, pipeline, results)

        node_sa_index = self._get_node_index(train_data, results)
        if node_sa_index is not None:
            node.rating = self._get_node_rating(node_sa_index)

        return results

    @staticmethod
    def _get_node_index(train_data: InputData, results: dict):
        total_index = None
        if NodeReplaceOperationAnalyze.__name__ in results.keys() and NodeDeletionAnalyze.__name__ in results.keys():
            task = train_data.task.task_type
            app_models, _ = OperationTypesRepository().suitable_operation(task_type=task)
            total_operations_number = len(app_models)

            replacement_candidates = results[NodeReplaceOperationAnalyze.__name__]
            candidates_for_replacement_number = len(
                [candidate for candidate in replacement_candidates if (1 - candidate) < 0])

            replacement_score = candidates_for_replacement_number / total_operations_number

            deletion_score = results[NodeDeletionAnalyze.__name__][0]

            total_index = (deletion_score / abs(deletion_score)) * replacement_score

        return total_index

    @staticmethod
    def _get_node_rating(total_index: float):
        rating = None
        if total_index <= -0.5:
            rating = 2
        elif -0.5 < total_index <= 0.0:
            rating = 1
        elif 0.0 < total_index <= 0.5:
            rating = 4
        elif 0.5 < total_index <= 1:
            rating = 3

        return rating

    def _save_results_to_json(self, node: Node, pipeline: Pipeline, results):
        node_id = pipeline.nodes.index(node)
        node_type = node.operation.operation_type
        result_file = join(self.path_to_save, f'{node_id}{node_type}_sa_results.json')
        with open(result_file, 'w', encoding='utf-8') as file:
            file.write(json.dumps(results, indent=4))

        self.log.info(f'Node Sensitivity analysis results were saved to {result_file}')


class NodeAnalyzeApproach(ABC):
    """Base class for analysis approach.

    Args:
        pipeline: :obj:`Pipeline` containing the analyzed :obj:`Node`
        train_data: data used for :obj:`Pipeline` training
        test_data: data used for :obj:`Pipeline` validation
        path_to_save: path to save results to. Default: ``~home/Fedot/sensitivity``
    """

    def __init__(self, pipeline: Pipeline, train_data, test_data: InputData,
                 requirements: SensitivityAnalysisRequirements = None,
                 path_to_save=None):
        self._pipeline = pipeline
        self._train_data = train_data
        self._test_data = test_data
        self._origin_metric = None
        self._requirements = \
            SensitivityAnalysisRequirements() if requirements is None else requirements

        self._path_to_save = \
            join(default_fedot_data_dir(), 'sensitivity', 'nodes_sensitivity') if path_to_save is None else path_to_save
        self.log = default_log(self)

        if not exists(self._path_to_save):
            makedirs(self._path_to_save)

    @abstractmethod
    def analyze(self, node: Node, **kwargs) -> Union[List[dict], List[float]]:
        """Creates the difference metric(scorer, index, etc) of the changed
        graph in relation to the original one

        Args:
            node: the sequence number of the node as in ``DFS`` result
        """
        pass

    @abstractmethod
    def sample(self, *args) -> Union[List[Pipeline], Pipeline]:
        """Changes the pipeline according to the approach"""
        pass

    def _compare_with_origin_by_metric(self, changed_pipeline: Pipeline) -> float:
        metric_id = MetricByTask.get_default_quality_metrics(self._train_data.task.task_type)[0]
        metric = MetricsRepository.metric_by_id(metric_id)

        if not self._origin_metric:
            self._origin_metric = self._get_metric_value(pipeline=self._pipeline, metric=metric)

        changed_pipeline_metric = self._get_metric_value(pipeline=changed_pipeline, metric=metric)

        return changed_pipeline_metric / self._origin_metric

    def _get_metric_value(self, pipeline: Pipeline, metric: Callable) -> float:
        pipeline.fit(self._train_data)
        predicted = pipeline.predict(self._test_data)
        metric_value = metric(true=self._test_data,
                              predicted=predicted)

        return metric_value


class NodeDeletionAnalyze(NodeAnalyzeApproach):
    def __init__(self, pipeline: Pipeline, train_data: InputData, test_data: InputData,
                 requirements: SensitivityAnalysisRequirements = None, path_to_save=None):
        super().__init__(pipeline, train_data, test_data, requirements,
                         path_to_save)

    def analyze(self, node: Node, **kwargs) -> Union[List[dict], List[float]]:
        """
        Args:
            node: :obj:`Node` object to analyze

        Returns:
            the ratio of modified pipeline score to origin score
        """

        if node is self._pipeline.root_node:
            # TODO or warning?
            return [1.0]
        else:
            shortened_pipeline = self.sample(node)
            if shortened_pipeline:
                loss = self._compare_with_origin_by_metric(shortened_pipeline)
                del shortened_pipeline
            else:
                loss = 1

            return [loss]

    def sample(self, node: Node):
        """
        Args:
            node: :obj:`Node` object to delete from :obj:`Pipeline` object
        Retuens:
            :obj:`Pipeline`: pipeline without node
        """

        pipeline_sample = deepcopy(self._pipeline)
        node_index_to_delete = self._pipeline.nodes.index(node)
        node_to_delete = pipeline_sample.nodes[node_index_to_delete]
        pipeline_sample.delete_node(node_to_delete)
        try:
            verify_pipeline(pipeline_sample)
        except ValueError as ex:
            self.log.info(f'Can not delete node. Deletion of this node leads to {ex}')
            return None

        return pipeline_sample

    def __str__(self):
        return 'NodeDeletionAnalyze'


class NodeReplaceOperationAnalyze(NodeAnalyzeApproach):
    """Replace node with operations available for the current task and evaluate the score difference
    """

    def __init__(self, pipeline: Pipeline, train_data: InputData, test_data: InputData,
                 requirements: SensitivityAnalysisRequirements = None, path_to_save=None):
        super().__init__(pipeline, train_data, test_data, requirements,
                         path_to_save)

    def analyze(self, node: Node, **kwargs) -> Union[List[dict], List[float]]:
        """
        Args:
            node: :obj:`Node` object to analyze

        Returns:
            the ratio of modified pipeline score to origin score
        """

        requirements: ReplacementAnalysisMetaParams = self._requirements.replacement_meta
        node_id = self._pipeline.nodes.index(node)
        samples = self.sample(node=node,
                              nodes_to_replace_to=requirements.nodes_to_replace_to,
                              number_of_random_operations=requirements.number_of_random_operations)

        loss_values = []
        new_nodes_types = []
        for sample_pipeline in samples:
            loss_per_sample = self._compare_with_origin_by_metric(sample_pipeline)
            loss_values.append(loss_per_sample)

            new_node = sample_pipeline.nodes[node_id]
            new_nodes_types.append(new_node.operation.operation_type)

        if self._requirements.visualization:
            self._visualize(x_values=new_nodes_types,
                            y_values=loss_values,
                            node_id=node_id)

        return loss_values

    def sample(self, node: Node,
               nodes_to_replace_to: Optional[List[Node]],
               number_of_random_operations: Optional[int] = None) -> Union[List[Pipeline], Pipeline]:
        """
        Args:
            node: :obj:`Node` object to replace
            nodes_to_replace_to: nodes provided for old_node replacement
            number_of_random_operations: number of replacement operations,
                if ``nodes_to_replace_to`` not provided
        Returns:
            Union[List[Pipeline], Pipeline]: sequence of :obj:`Pipeline` objects with new operations instead of old one
        """

        if not nodes_to_replace_to:
            node_type = type(node)
            nodes_to_replace_to = self._node_generation(node_type=node_type,
                                                        number_of_operations=number_of_random_operations)

        samples = list()
        for replacing_node in nodes_to_replace_to:
            sample_pipeline = deepcopy(self._pipeline)
            replaced_node_index = self._pipeline.nodes.index(node)
            replaced_node = sample_pipeline.nodes[replaced_node_index]
            sample_pipeline.update_node(old_node=replaced_node,
                                        new_node=replacing_node)
            samples.append(sample_pipeline)

        return samples

    def _node_generation(self, node_type: Type[Node],
                         number_of_operations=None) -> List[Node]:
        task = self._train_data.task.task_type
        # Get models
        app_models, _ = OperationTypesRepository().suitable_operation(task_type=task)
        # Get data operations for such task
        app_data_operations = OperationTypesRepository('data_operation').suitable_operation(
            task_type=task)

        # Unit two lists
        app_operations = app_models
        if number_of_operations:
            random_operations = random.sample(app_operations, number_of_operations)
        else:
            random_operations = app_operations

        nodes = []
        for operation in random_operations:
            nodes.append(node_type(operation_type=operation))

        return nodes

    def _visualize(self, x_values, y_values: list, node_id: int):
        data = zip(x_values, [(y - 1) for y in y_values])  # 3
        original_operation_type = self._pipeline.nodes[node_id].operation.operation_type

        sorted_data = sorted(data, key=lambda tup: tup[1])
        x_values = [x[0] for x in sorted_data]
        y_values = [y[1] for y in sorted_data]
        colors = ['r' if y < 0 else 'g' for y in y_values]
        fig, ax = plt.subplots(figsize=(15, 10))
        ax.bar(x_values, y_values, width=1.0, edgecolor='black', bottom=1,
               color=colors)
        ax.hlines(1, -1, len(x_values) + 1, linestyle='--')
        ax.set_xticklabels(x_values, rotation=45)
        plt.xlabel('iteration')
        plt.ylabel('quality (changed_pipeline_metric/original_metric) - 1')

        if original_operation_type in x_values:
            original_operation_index = x_values.index(original_operation_type)
            plt.gca().get_xticklabels()[original_operation_index].set_color('red')

        file_name = f'{self._pipeline.nodes[node_id].operation.operation_type}_id_{node_id}_replacement.jpg'
        result_file = join(self._path_to_save, file_name)
        plt.savefig(result_file)
        self.log.info(f'NodeReplacementAnalysis for '
                         f'{original_operation_type}(index:{node_id}) was saved to {result_file}')

    def __str__(self):
        return 'NodeReplaceOperationAnalyze'



from random import choice
from typing import Optional, Iterable

from fedot.core.pipelines.pipeline import Pipeline


class OperationsFilter:
    def satisfies(self, pipeline: Optional[Pipeline]) -> bool:
        """ Checks if all operations in a Pipeline satisify this filter. """
        return True

    def sample(self) -> str:
        """ Samples some operation that satisfies this filter. """
        raise NotImplementedError()


class WhitelistOperationsFilter(OperationsFilter):
    """ Simple OperationsFilter implementation based on two lists:
    one for all available operations, another for sampling operations. """

    def __init__(self, available_operations: Iterable[str], available_task_operations: Optional[Iterable[str]] = None):
        super().__init__()
        self._whitelist = tuple(available_operations)
        self._choice_operations = tuple(available_task_operations) if available_task_operations else self._whitelist

    def satisfies(self, pipeline: Optional[Pipeline]) -> bool:
        def node_ok(node):
            return node.operation.operation_type in self._whitelist

        return pipeline and all(map(node_ok, pipeline.nodes))

    def sample(self) -> str:
        return choice(self._choice_operations)



import warnings
from typing import Optional

from fedot.core.data.data import InputData, OutputData
from fedot.core.operations.evaluation.evaluation_interfaces import EvaluationStrategy
from fedot.core.operations.operation_parameters import OperationParameters

warnings.filterwarnings("ignore", category=UserWarning)


class DataSourceStrategy(EvaluationStrategy):

    def __init__(self, operation_type: str, params: Optional[OperationParameters] = None):
        super().__init__(operation_type, params)

    def fit(self, train_data: InputData):
        return object()

    def predict(self, trained_operation, predict_data: InputData) -> OutputData:
        return OutputData(idx=predict_data.idx, features=predict_data.features, task=predict_data.task,
                          data_type=predict_data.data_type, target=predict_data.target, predict=predict_data.features)

    def _convert_to_operation(self, operation_type: str):
        return object()



from os.path import join

from examples.advanced.sensitivity_analysis.dataset_access import get_scoring_data
from examples.advanced.sensitivity_analysis.pipelines_access import get_three_depth_manual_class_pipeline
from fedot.core.utils import default_fedot_data_dir
from fedot.sensitivity.node_sa_approaches import NodeDeletionAnalyze, NodeReplaceOperationAnalyze
from fedot.sensitivity.operations_hp_sensitivity.multi_operations_sensitivity import MultiOperationsHPAnalyze
from fedot.sensitivity.pipeline_sensitivity_facade import PipelineSensitivityAnalysis
from fedot.sensitivity.sa_requirements import SensitivityAnalysisRequirements


def run_analysis(pipeline, train_data, test_data):
    sa_requirements = SensitivityAnalysisRequirements(visualization=True,
                                                      is_save_results_to_json=True)
    approaches = [NodeDeletionAnalyze, NodeReplaceOperationAnalyze, MultiOperationsHPAnalyze]
    result_path = join(default_fedot_data_dir(), 'sensitivity', f'{PipelineSensitivityAnalysis.__name__}')

    PipelineSensitivityAnalysis(pipeline=pipeline, train_data=train_data,
                             test_data=test_data, approaches=approaches,
                             requirements=sa_requirements, path_to_save=result_path).analyze()


if __name__ == '__main__':
    pipeline = get_three_depth_manual_class_pipeline()
    train_data, test_data = get_scoring_data()

    run_analysis(pipeline=pipeline, train_data=train_data, test_data=test_data)



from typing import Union

from fedot.core.caching.pipelines_cache_db import OperationsCacheDB
from fedot.core.caching.preprocessing_cache_db import PreprocessingCacheDB
from fedot.core.log import default_log
from fedot.core.utilities.singleton_meta import SingletonMeta


class BaseCache(metaclass=SingletonMeta):
    """
    Stores/loads data to increase performance.

    :param cache_db: specific DB for specific data
    """

    def __init__(self, cache_db: Union[OperationsCacheDB, PreprocessingCacheDB]):
        self._db = cache_db
        self.log = default_log(__name__)

    @property
    def effectiveness_ratio(self):
        """
        Returns percent of how many elements were loaded instead of computing.
        """
        if self._db.use_stats:
            #  Result order corresponds to the order in self.db._effectiveness_keys
            eff_dct = {}
            returned_eff = self._db.get_effectiveness()
            for key, hit, total in zip(self._db.get_effectiveness_keys()[::2], returned_eff[::2], returned_eff[1::2]):
                key = key.split('_')[0]
                eff_dct[key] = round(hit / total, 3) if total else 0.
            return eff_dct

    def reset(self):
        """
        Drops all scores from working table and resets efficiency table values to zero.
        """
        self._db.reset()

    def __len__(self):
        return len(self._db)



import pandas as pd

from fedot.api.main import Fedot
from fedot.core.utils import fedot_project_root


def run_api_explain_example(visualization=False, timeout=None):
    train_data = pd.read_csv(f'{fedot_project_root()}/cases/data/cancer/cancer_train.csv', index_col=0)
    figure_path = 'api_explain_example.png'

    # Feature and class names for visualization
    feature_names = train_data.columns.tolist()
    target_name = feature_names.pop()
    target = train_data[target_name]
    class_names = target.unique().astype(str).tolist()

    # Building simple pipeline
    model = Fedot(problem='classification', timeout=timeout)
    model.fit(features=train_data, target=target_name, predefined_model='rf')

    # Current pipeline explaining
    explainer = model.explain(
        method='surrogate_dt', visualization=visualization,
        # The following parameters are only used if visualize == True:
        save_path=figure_path, dpi=200, feature_names=feature_names,
        class_names=class_names,
        precision=6
    )

    return explainer


if __name__ == '__main__':
    run_api_explain_example(visualization=True, timeout=5)



import random

from hyperopt.pyll.stochastic import sample as hp_sample

from fedot.core.log import default_log
from fedot.core.pipelines.tuning.search_space import SearchSpace


class ParametersChanger:
    """
    Class for the hyperparameters changing in the operation

    :param operation_name: name of operation to get hyperparameters for
    :param current_params: current parameters value
    """

    def __init__(self, operation_name, current_params):
        self.operation_name = operation_name
        self.current_params = current_params
        self.logger = default_log(prefix='ParametersChangerLog')

    def get_new_operation_params(self):
        """ Function return a dictionary with new parameters values """

        # Get available parameters for operation
        params_list = SearchSpace().get_operation_parameter_range(self.operation_name)

        if params_list is None:
            params_dict = None
        else:
            # Get new values for all parameters
            params_dict = self.new_params_dict(params_list)

        return params_dict

    def new_params_dict(self, params_list):
        """ Change values of hyperparameters by different ways

        :param params_list: list with hyperparameters names
        """

        _change_by_name = {'lagged': {'window_size': self._incremental_change},
                           'sparse_lagged': {'window_size': self._incremental_change}}

        params_dict = {}
        for parameter_name in params_list:
            # Get current value of the parameter
            current_value = self._get_current_parameter_value(parameter_name)

            # Perform parameter value change using appropriate function

            operation_dict = _change_by_name.get(self.operation_name)
            if operation_dict is not None:
                func = operation_dict.get(parameter_name)
            else:
                # Default changes perform with random choice
                func = self._random_change
            if func is None:
                func = self._random_change
            parameters = {'operation_name': self.operation_name,
                          'current_value': current_value}

            param_value = func(parameter_name, **parameters)
            params_dict.update(param_value)

        return params_dict

    def _get_current_parameter_value(self, parameter_name):

        if isinstance(self.current_params, str):
            # TODO 'default_params' - need to process
            current_value = None
        else:
            # Dictionary with parameters
            try:
                current_value = self.current_params.get(parameter_name)
            except Exception as exec:
                self.logger.warning(f'The following error occurred during the hyperparameter configuration.{exec}')
                current_value = None

        return current_value

    @staticmethod
    def _random_change(parameter_name, **kwargs):
        """ Randomly selects a parameter value from a specified range """

        space = SearchSpace().get_operation_parameter_range(operation_name=kwargs['operation_name'],
                                                            parameter_name=parameter_name,
                                                            label=parameter_name)
        # Randomly choose new value
        new_value = hp_sample(space)
        return {parameter_name: new_value}

    @staticmethod
    def _incremental_change(parameter_name, **kwargs):
        """ Next to the current value, the normally distributed new value is set aside """
        # TODO add the ability to limit the boundaries of the params ranges
        sigma = kwargs['current_value'] * 0.3
        new_value = random.normalvariate(kwargs['current_value'], sigma)
        return {parameter_name: new_value}



from examples.advanced.sensitivity_analysis.dataset_access import get_scoring_data
from examples.advanced.sensitivity_analysis.pipelines_access import get_three_depth_manual_class_pipeline
from fedot.core.data.data_split import train_test_data_setup
from fedot.sensitivity.deletion_methods.multi_times_analysis import MultiTimesAnalyze


def run_mta_analysis(pipeline, train_data, test_data, valid_data):
    size_reduction_ratio = MultiTimesAnalyze(pipeline=pipeline, train_data=train_data, test_data=test_data,
                                             valid_data=valid_data, case_name='scoring_mta_experiment'). \
        analyze(visualization=True)
    print(f'The number of deleted nodes to the original pipeline length is {size_reduction_ratio}')


if __name__ == '__main__':
    pipeline = get_three_depth_manual_class_pipeline()
    train_data, test_data = get_scoring_data()
    test_data, valid_data = train_test_data_setup(test_data, split_ratio=0.5)

    run_mta_analysis(pipeline, train_data, test_data, valid_data)



from typing import Union

# imports are required for the eval
from fedot.core.repository.dataset_types import *
from fedot.core.repository.tasks import *


def read_field(source: dict, field_name: str, default: list):
    """ Function for reading field in the dictionary

    :param source: dictionary with information
    :param field_name: name of the field for searching for in it
    :param default: default list if field_name is not in the source dict keys

    :return : list with field values
    """
    if field_name in source.keys():
        field_value = source[field_name]
        if isinstance(field_value, str):
            return eval_field_str(field_value)
        else:
            return field_value
    else:
        return default


def eval_field_str(field_value) -> Union[List[DataTypesEnum],
                                         List[TaskTypesEnum]]:
    # TODO add docstring
    return eval(field_value)


def eval_strategy_str(field_value):
    # TODO add docstring
    namespace = field_value[0]
    exec(f'from {namespace} import {field_value[1]}')
    return eval(field_value[1])



from abc import abstractmethod
from typing import List

import numpy as np
import pandas as pd

from fedot.core.constants import FRACTION_OF_UNIQUE_VALUES

ALLOWED_NAN_PERCENT = 0.9


class DataDetector:

    @staticmethod
    @abstractmethod
    def prepare_multimodal_data(dataframe: pd.DataFrame, columns: List[str]) -> dict:
        raise NotImplementedError()

    @staticmethod
    @abstractmethod
    def new_key_name(data_part_key: str) -> str:
        raise NotImplementedError()


class TextDataDetector(DataDetector):
    """
    Class for detecting text data during its import.
    """

    def define_text_columns(self, data_frame: pd.DataFrame) -> List[str]:
        """
        :param data_frame: pandas dataframe with data
        :return: list of text columns' names
        """
        text_columns = []
        for column_name in data_frame.columns:
            if self._column_contains_text(data_frame[column_name]):
                text_columns.append(column_name)
        return text_columns

    @staticmethod
    def is_full_of_nans(text_data: np.array) -> bool:
        if np.sum(pd.isna(text_data)) / len(text_data) > ALLOWED_NAN_PERCENT:
            return True
        return False

    @staticmethod
    def prepare_multimodal_data(dataframe: pd.DataFrame, columns: List[str]) -> dict:
        """ Prepares MultiModal text data in a form of dictionary

        :param dataframe: pandas DataFrame to process
        :param columns: list of text columns' names

        :return multimodal_text_data: dictionary with numpy arrays of text data
        """
        multi_modal_text_data = {}

        for column_name in columns:
            text_feature = np.array(dataframe[column_name])
            multi_modal_text_data.update({column_name: text_feature})

        return multi_modal_text_data

    @staticmethod
    def new_key_name(data_part_key: str) -> str:
        return f'data_source_text/{data_part_key}'

    def _column_contains_text(self, column: pd.Series) -> bool:
        """
        Column contains text if:
        1. it's not float or float compatible
        (e.g. ['1.2', '2.3', '3.4', ...] is float too)
        2. fraction of unique values (except nans) is more than 0.95

        :param column: pandas series with data
        :return: True if column contains text
        """
        if column.dtype == object and not self._is_float_compatible(column):
            unique_num = len(column.unique())
            nan_num = pd.isna(column).sum()
            return unique_num / len(column) > FRACTION_OF_UNIQUE_VALUES if nan_num == 0 \
                else (unique_num - 1) / (len(column) - nan_num) > FRACTION_OF_UNIQUE_VALUES
        return False

    @staticmethod
    def _is_float_compatible(column: pd.Series) -> bool:
        """
        :param column: pandas series with data
        :return: True if column contains only float or nan values
        """
        nans_number = column.isna().sum()
        converted_column = pd.to_numeric(column, errors='coerce')
        result_nans_number = converted_column.isna().sum()
        failed_objects_number = result_nans_number - nans_number
        non_nan_all_objects_number = len(column) - nans_number
        failed_ratio = failed_objects_number / non_nan_all_objects_number
        return failed_ratio < 0.5


class TimeSeriesDataDetector(DataDetector):
    """
    Class for detecting time series data during its import.
    """

    @staticmethod
    def prepare_multimodal_data(dataframe: pd.DataFrame, columns: List[str]) -> dict:
        """ Prepares MultiModal data for time series forecasting task in a form of dictionary

        :param dataframe: pandas DataFrame to process
        :param columns: column names, which should be used as features in forecasting

        :return multi_modal_ts_data: dictionary with numpy arrays
        """
        multi_modal_ts_data = {}
        for column_name in columns:
            feature_ts = np.array(dataframe[column_name])

            # Will be the same
            multi_modal_ts_data.update({column_name: feature_ts})

        return multi_modal_ts_data

    @staticmethod
    def new_key_name(data_part_key: str) -> str:
        if data_part_key == 'idx':
            return 'idx'
        return f'data_source_ts/{data_part_key}'



import pickle
import sqlite3
from contextlib import closing
from typing import Dict, Optional, Tuple

from fedot.core.caching.base_cache_db import BaseCacheDB
from fedot.core.operations.evaluation.operation_implementations.data_operations.categorical_encoders import (
    OneHotEncodingImplementation
)
from fedot.core.operations.evaluation.operation_implementations.data_operations.sklearn_transformations import (
    ImputationImplementation
)
from fedot.preprocessing.preprocessing import DataPreprocessor


class PreprocessingCacheDB(BaseCacheDB):
    """
    Database for `PreprocessingCache` class.
    Includes low-level idea of caching pipeline preprocessor items using relational database.

    :param cache_folder: path to the place where cache files should be stored.
    """

    def __init__(self, cache_folder: Optional[str] = None):
        super().__init__('preprocessors', cache_folder, False, ['preprocessors_hit', 'preprocessors_total'])
        self._init_db()

    def get_preprocessor(self, uid: str) -> Optional[Tuple[
        Dict[str, OneHotEncodingImplementation], Dict[str, ImputationImplementation]
    ]]:
        """
        Tries to return both data processors, None if is not found

        :param uid: uid of preproccesor to be loaded

        :return matched: pair of data processors (encoder, imputer) or None
        """
        with closing(sqlite3.connect(self.db_path)) as conn:
            with conn:
                cur = conn.cursor()
                cur.execute(f'SELECT encoder, imputer FROM {self._main_table} WHERE id = ?;', [uid])
                matched = cur.fetchone()
                is_loaded = False
                if matched is not None:
                    matched = tuple([pickle.loads(matched[i]) for i in range(2)])
                    is_loaded = True
                if self.use_stats:
                    if is_loaded:
                        self._inc_eff(cur, 'preprocessors_hit')
                    self._inc_eff(cur, 'preprocessors_total')
                return matched

    def add_preprocessor(self, uid: str, value: DataPreprocessor):
        """
        Adds preprocessor score to DB table vid its uid.

        :param uid: unique preprocessor identificator
        :param value: the preprocessor itself
        """
        with closing(sqlite3.connect(self.db_path)) as conn:
            with conn:
                cur = conn.cursor()
                pickled_encoder = sqlite3.Binary(pickle.dumps(value.features_encoders, pickle.HIGHEST_PROTOCOL))
                pickled_imputer = sqlite3.Binary(pickle.dumps(value.features_imputers, pickle.HIGHEST_PROTOCOL))
                cur.execute(f'INSERT OR IGNORE INTO {self._main_table} VALUES (?, ?, ?);',
                            [uid, pickled_encoder, pickled_imputer])

    def _init_db(self):
        """
        Initializes DB working table.
        """
        with closing(sqlite3.connect(self.db_path)) as conn:
            with conn:
                cur = conn.cursor()
                cur.execute((
                    f'CREATE TABLE IF NOT EXISTS {self._main_table} ('
                    'id TEXT PRIMARY KEY,'
                    'encoder BLOB,'
                    'imputer BLOB'
                    ');'
                ))



import logging
from copy import deepcopy

import pandas as pd

from fedot.api.main import Fedot
from fedot.core.operations.atomized_model import AtomizedModel
from fedot.core.pipelines.node import PrimaryNode, SecondaryNode
from fedot.core.pipelines.pipeline import Pipeline
from fedot.core.pipelines.pipeline_builder import PipelineBuilder
from fedot.core.utils import fedot_project_root


def run_additional_learning_example():
    train_data_path = f'{fedot_project_root()}/cases/data/scoring/scoring_train.csv'
    test_data_path = f'{fedot_project_root()}/cases/data/scoring/scoring_test.csv'

    train_data = pd.read_csv(train_data_path)
    test_data = pd.read_csv(test_data_path)
    test_data_target = test_data['target']
    del test_data['target']

    problem = 'classification'

    auto_model = Fedot(problem=problem, seed=42, timeout=5, preset='best_quality',
                       initial_assumption=PipelineBuilder().add_node('scaling').add_node('logit').to_pipeline())

    auto_model.fit(features=deepcopy(train_data.head(1000)), target='target')
    auto_model.predict_proba(features=deepcopy(test_data))
    print('auto_model', auto_model.get_metrics(target=deepcopy(test_data_target)))

    prev_model = auto_model.current_pipeline
    prev_model.show()

    prev_model.unfit()
    atomized_model = Pipeline(
        SecondaryNode(operation_type=AtomizedModel(prev_model), nodes_from=[PrimaryNode('scaling')]))
    non_atomized_model = deepcopy(prev_model)

    train_data = train_data.head(5000)
    timeout = 1

    auto_model_from_atomized = Fedot(problem=problem, seed=42, preset='best_quality', timeout=timeout,
                                     logging_level=logging.INFO,
                                     initial_assumption=atomized_model)
    auto_model_from_atomized.fit(features=deepcopy(train_data), target='target')
    auto_model_from_atomized.predict_proba(features=deepcopy(test_data))
    auto_model_from_atomized.current_pipeline.show()
    print('auto_model_from_atomized', auto_model_from_atomized.get_metrics(deepcopy(test_data_target)))

    auto_model_from_pipeline = Fedot(problem=problem, seed=42, preset='best_quality', timeout=timeout,
                                     logging_level=logging.INFO,
                                     initial_assumption=non_atomized_model)
    auto_model_from_pipeline.fit(features=deepcopy(train_data), target='target')
    auto_model_from_pipeline.predict_proba(features=deepcopy(test_data))
    auto_model_from_pipeline.current_pipeline.show()
    print('auto_model_from_pipeline', auto_model_from_pipeline.get_metrics(deepcopy(test_data_target)))


if __name__ == '__main__':
    run_additional_learning_example()



from typing import Union, Optional, Iterable, List
from uuid import uuid4

from fedot.core.dag.graph_node import GraphNode
from fedot.core.utilities.data_structures import UniqueList


class LinkedGraphNode(GraphNode):
    """Class for node definition in the directed graph structure
    that directly stores its parent nodes.

    Args:
        nodes_from: parent nodes which information comes from
        content: ``dict`` for the content in the node

    Notes:
        The possible parameters are:
            - ``name`` - name (str) or object that performs actions in this node
            - ``params`` - dictionary with additional information that is used by
                    the object in the ``name`` field (e.g. hyperparameters values)
    """

    def __init__(self, content: Union[dict, str],
                 nodes_from: Optional[Iterable['LinkedGraphNode']] = None):
        # Wrap string into dict if it is necessary
        if isinstance(content, str):
            content = {'name': content}

        self.content = content
        self._nodes_from = UniqueList(nodes_from or ())

        super().__init__()

    @property
    def nodes_from(self) -> List['LinkedGraphNode']:
        return self._nodes_from

    @nodes_from.setter
    def nodes_from(self, nodes: Optional[Iterable['LinkedGraphNode']]):
        self._nodes_from = UniqueList(nodes)

    @property
    def name(self) -> str:
        return str(self.content.get('name'))

    def __hash__(self) -> int:
        return hash(self.uid)

    def __str__(self) -> str:
        return self.name or self.uid

    def __repr__(self) -> str:
        return self.__str__()

    def description(self) -> str:
        node_operation = self.content.get('name')
        if node_operation is None:
            raise ValueError('Unknown content in the node: expected content[`name`], got None')
        params = self.content.get('params')
        # TODO: possibly unify with __repr__ & don't duplicate Operation.description
        if not params :
            node_label = f'n_{node_operation}'
        elif isinstance(node_operation, str):
            # If there is a string: name of operation (as in json repository)
            node_label = f'n_{node_operation}_{params}'
        else:
            # If instance of Operation is placed in 'name'
            node_label = node_operation.description(params)
        return node_label



from abc import abstractmethod


class Explainer:
    """
    An abstract class for various explanation methods.
    """
    def __init__(self, model):
        self.model = model

    @abstractmethod
    def explain(self, *args, **kwargs):
        raise NotImplementedError

    @abstractmethod
    def visualize(self, *args, **kwargs):
        raise NotImplementedError



import glob
import os
from copy import copy, deepcopy
from dataclasses import dataclass, field
from typing import List, Optional, Tuple, Union

import numpy as np
import pandas as pd

from fedot.utilities.requirements_notificator import warn_requirement

try:
    import cv2
except ModuleNotFoundError:
    warn_requirement('opencv-python')
    cv2 = None

from fedot.core.data.array_utilities import atleast_2d
from fedot.core.data.load_data import JSONBatchLoader, TextBatchLoader
from fedot.core.data.supplementary_data import SupplementaryData
from fedot.core.repository.dataset_types import DataTypesEnum
from fedot.core.repository.tasks import Task, TaskTypesEnum


@dataclass
class Data:
    """
    Base Data type class
    """

    idx: np.array
    task: Task
    data_type: DataTypesEnum
    features: np.array
    target: Optional[np.array] = None

    # Object with supplementary info
    supplementary_data: SupplementaryData = field(default_factory=SupplementaryData)

    @staticmethod
    def from_csv(file_path=None,
                 delimiter=',',
                 task: Task = Task(TaskTypesEnum.classification),
                 data_type: DataTypesEnum = DataTypesEnum.table,
                 columns_to_drop: Optional[List] = None,
                 target_columns: Union[str, List] = '',
                 index_col: Optional[Union[str, int]] = 0):
        """Import data from ``csv``

        Args:
            file_path: the path to the ``CSV`` with data
            columns_to_drop: the names of columns that should be dropped
            delimiter: the delimiter to separate the columns
            task: the :obj:`Task` that should be solved with data
            data_type: the type of data interpretation
            target_columns: name of target column (last column if empty and no target if ``None``)
            index_col: column name or index to use as the :obj:`Data.idx`;\n
                if ``None`` then arrange new unique index

        Returns:
            data
        """

        data_frame = pd.read_csv(file_path, sep=delimiter, index_col=index_col)
        if columns_to_drop:
            data_frame = data_frame.drop(columns_to_drop, axis=1)

        idx = data_frame.index.to_numpy()
        features, target = process_target_and_features(data_frame, target_columns)

        return InputData(idx=idx, features=features, target=target, task=task, data_type=data_type)

    @staticmethod
    def from_csv_time_series(task: Task,
                             file_path=None,
                             delimiter=',',
                             is_predict=False,
                             target_column: Optional[str] = ''):
        df = pd.read_csv(file_path, sep=delimiter)

        idx = get_indices_from_file(df, file_path)

        if target_column is not None:
            time_series = np.array(df[target_column])
        else:
            time_series = np.array(df[df.columns[-1]])

        if is_predict:
            # Prepare data for prediction
            len_forecast = task.task_params.forecast_length

            start_forecast = len(time_series)
            end_forecast = start_forecast + len_forecast
            input_data = InputData(idx=np.arange(start_forecast, end_forecast),
                                   features=time_series,
                                   target=None,
                                   task=task,
                                   data_type=DataTypesEnum.ts)
        else:
            # Prepare InputData for train the pipeline
            input_data = InputData(idx=idx,
                                   features=time_series,
                                   target=time_series,
                                   task=task,
                                   data_type=DataTypesEnum.ts)

        return input_data

    @staticmethod
    def from_csv_multi_time_series(task: Task,
                                   file_path=None,
                                   delimiter=',',
                                   is_predict=False,
                                   columns_to_use: Optional[list] = None,
                                   target_column: Optional[str] = ''):
        """
        Forms :obj:`InputData` of ``multi_ts`` type from columns of different variant of the same variable

        Args:
            task: the :obj:`Task` that should be solved with data
            file_path: path to csv file
            delimiter: delimiter for pandas df
            is_predict: is preparing for fit or predict stage
            columns_to_use: ``list`` with names of columns of different variant of the same variable
            target_column: ``string`` with name of target column, used for predict stage

        Returns:
            data
        """

        df = pd.read_csv(file_path, sep=delimiter)

        idx = get_indices_from_file(df, file_path)
        if columns_to_use is not None:
            actual_df = df[columns_to_use]
            multi_time_series = actual_df.to_numpy()
        else:
            multi_time_series = df.to_numpy()

        if is_predict:
            # Prepare data for prediction
            len_forecast = task.task_params.forecast_length
            if target_column is not None:
                time_series = np.array(df[target_column])
            else:
                time_series = np.array(df[df.columns[-1]])
            start_forecast = multi_time_series.shape[0]
            end_forecast = start_forecast + len_forecast
            input_data = InputData(idx=np.arange(start_forecast, end_forecast),
                                   features=time_series,
                                   target=None,
                                   task=task,
                                   data_type=DataTypesEnum.multi_ts)
        else:
            # Prepare InputData for train the pipeline
            input_data = InputData(idx=idx,
                                   features=multi_time_series,
                                   target=multi_time_series,
                                   task=task,
                                   data_type=DataTypesEnum.multi_ts)

        return input_data

    @staticmethod
    def from_image(images: Union[str, np.ndarray] = None,
                   labels: Union[str, np.ndarray] = None,
                   task: Task = Task(TaskTypesEnum.classification),
                   target_size: Optional[Tuple[int, int]] = None):
        """Input data from Image

        Args:
            images: the path to the directory with image data in ``np.ndarray`` format
                or array in ``np.ndarray`` format
            labels: the path to the directory with image labels in ``np.ndarray`` format
                or array in ``np.ndarray`` format
            task: the :obj:`Task` that should be solved with data
            target_size: size for the images resizing (if necessary)

        Returns:
            data
        """

        features = images
        target = labels

        if type(images) is str:
            # if upload from path
            if '*.jpeg' in images:
                # upload from folder of images
                path = images
                images_list = []
                for file_path in glob.glob(path):
                    if target_size is not None:
                        img = _resize_image(file_path, target_size)
                        images_list.append(img)
                    else:
                        raise ValueError('Set target_size for images')
                features = np.asarray(images_list)
                target = labels
            else:
                # upload from array
                features = np.load(images)
                target = np.load(labels)
                # add channels if None
                if len(features.shape) == 3:
                    features = np.expand_dims(features, -1)

        idx = np.arange(0, len(features))

        return InputData(idx=idx, features=features, target=target, task=task, data_type=DataTypesEnum.image)

    @staticmethod
    def from_text_meta_file(meta_file_path: str = None,
                            label: str = 'label',
                            task: Task = Task(TaskTypesEnum.classification),
                            data_type: DataTypesEnum = DataTypesEnum.text):

        if os.path.isdir(meta_file_path):
            raise ValueError("""CSV file expected but got directory""")

        df_text = pd.read_csv(meta_file_path)
        df_text = df_text.sample(frac=1).reset_index(drop=True)
        messages = df_text['text'].astype('U').tolist()

        features = np.array(messages)
        target = np.array(df_text[label]).reshape(-1, 1)
        idx = [index for index in range(len(target))]

        return InputData(idx=idx, features=features,
                         target=target, task=task, data_type=data_type)

    @staticmethod
    def from_text_files(files_path: str,
                        label: str = 'label',
                        task: Task = Task(TaskTypesEnum.classification),
                        data_type: DataTypesEnum = DataTypesEnum.text):

        if os.path.isfile(files_path):
            raise ValueError("""Path to the directory expected but got file""")

        df_text = TextBatchLoader(path=files_path).extract()

        features = np.array(df_text['text'])
        target = np.array(df_text[label]).reshape(-1, 1)
        idx = [index for index in range(len(target))]

        return InputData(idx=idx, features=features,
                         target=target, task=task, data_type=data_type)

    @staticmethod
    def from_json_files(files_path: str,
                        fields_to_use: List,
                        label: str = 'label',
                        task: Task = Task(TaskTypesEnum.classification),
                        data_type: DataTypesEnum = DataTypesEnum.table,
                        export_to_meta=False, is_multilabel=False, shuffle=True) -> 'InputData':
        """Generates InputData from the set of ``JSON`` files with different fields

        Args:
            files_path: path the folder with ``json`` files
            fields_to_use: ``list`` of fields that will be considered as a features
            label: name of field with target variable
            task: :obj:`Task` to solve
            data_type: data type in fields (as well as type for obtained :obj:`InputData`)
            export_to_meta: combine extracted field and save to ``CSV``
            is_multilabel: if ``True``, creates multilabel target
            shuffle: if ``True``, shuffles data

        Returns:
            combined dataset
        """

        if os.path.isfile(files_path):
            raise ValueError("""Path to the directory expected but got file""")

        df_data = JSONBatchLoader(path=files_path, label=label, fields_to_use=fields_to_use,
                                  shuffle=shuffle).extract(export_to_meta)

        if len(fields_to_use) > 1:
            fields_to_combine = []
            for field in fields_to_use:
                fields_to_combine.append(np.array(df_data[field]))
                # Unite if the element of text data is divided into strings
                if isinstance(df_data[field][0], list):
                    df_data[field] = [' '.join(piece) for piece in df_data[field]]

            features = np.column_stack(tuple(fields_to_combine))
        else:
            field = df_data[fields_to_use[0]]
            # process field with nested list
            if isinstance(field[0], list):
                field = [' '.join(piece) for piece in field]
            features = np.array(field)

        if is_multilabel:
            target = df_data[label]
            classes = set()
            for el in target:
                for label in el:
                    classes.add(label)
            count_classes = list(sorted(classes))
            multilabel_target = np.zeros((len(features), len(count_classes)))

            for i in range(len(target)):
                for el in target[i]:
                    multilabel_target[i][count_classes.index(el)] = 1
            target = multilabel_target
        else:
            target = np.array(df_data[label])

        idx = [index for index in range(len(target))]

        return InputData(idx=idx, features=features,
                         target=target, task=task, data_type=data_type)

    def to_csv(self, path_to_save):
        dataframe = pd.DataFrame(data=self.features, index=self.idx)
        if self.target is not None:
            dataframe['target'] = self.target
        dataframe.to_csv(path_to_save)


@dataclass
class InputData(Data):
    """Data class for input data for the nodes
    """

    @property
    def num_classes(self) -> Optional[int]:
        """Returns number of classes that are present in the target.
        NB: if some labels are not present in this data, then
        number of classes can be less than in the full dataset!"""
        unique_values = self.class_labels
        return len(unique_values) if unique_values is not None else None

    @property
    def class_labels(self) -> Optional[int]:
        """Returns unique class labels that are present in the target"""
        if self.task.task_type == TaskTypesEnum.classification and self.target is not None:
            return np.unique(self.target)
        else:
            return None

    def subset_range(self, start: int, end: int):
        if not (0 <= start <= end <= len(self.idx)):
            raise ValueError('Incorrect boundaries for subset')
        new_features = None
        if self.features is not None:
            new_features = self.features[start:end + 1]
        return InputData(idx=self.idx[start:end + 1], features=new_features,
                         target=self.target[start:end + 1],
                         task=self.task, data_type=self.data_type)

    def subset_indices(self, selected_idx: List):
        """Get subset from :obj:`InputData` to extract all items with specified indices

        Args:
            selected_idx: ``list`` of indices for extraction

        Returns:
            :obj:`InputData`
        """

        idx_list = [str(i) for i in self.idx]

        # extractions of row number for each existing index from selected_idx
        row_nums = [idx_list.index(str(selected_ind)) for selected_ind in selected_idx
                    if str(selected_ind) in idx_list]
        new_features = None

        if self.features is not None:
            new_features = self.features[row_nums]
        return InputData(idx=np.asarray(self.idx)[row_nums], features=new_features,
                         target=self.target[row_nums],
                         task=self.task, data_type=self.data_type)

    def subset_features(self, features_ids: list):
        """Return new :obj:`InputData` with subset of features based on ``features_ids`` list
        """

        subsample_features = self.features[:, features_ids]
        subsample_input = InputData(features=subsample_features,
                                    data_type=self.data_type,
                                    target=self.target, task=self.task,
                                    idx=self.idx,
                                    supplementary_data=self.supplementary_data)

        return subsample_input

    def shuffle(self):
        """Shuffles features and target if possible
        """

        if self.data_type in (DataTypesEnum.table, DataTypesEnum.image, DataTypesEnum.text):
            shuffled_ind = np.random.permutation(len(self.features))
            idx, features, target = np.asarray(self.idx)[shuffled_ind], self.features[shuffled_ind], self.target[
                shuffled_ind]
            self.idx = idx
            self.features = features
            self.target = target
        else:
            pass

    def convert_non_int_indexes_for_fit(self, pipeline):
        """Conversion non ``int`` (``datetime``, ``string``, etc) indexes in ``integer`` form on the fit stage
        """

        copied_data = deepcopy(self)
        is_timestamp = isinstance(copied_data.idx[0], pd._libs.tslibs.timestamps.Timestamp)
        is_numpy_datetime = isinstance(copied_data.idx[0], np.datetime64)
        # if fit stage- just creating range of integers
        if is_timestamp or is_numpy_datetime:
            copied_data.supplementary_data.non_int_idx = copy(copied_data.idx)
            copied_data.idx = np.array(range(len(copied_data.idx)))

            last_idx_time = copied_data.supplementary_data.non_int_idx[-1]
            pre_last_time = copied_data.supplementary_data.non_int_idx[-2]

            pipeline.last_idx_int = copied_data.idx[-1]
            pipeline.last_idx_dt = last_idx_time
            pipeline.period = last_idx_time - pre_last_time
        elif not isinstance(copied_data.idx[0], (int, np.int32, np.int64)):
            copied_data.supplementary_data.non_int_idx = copy(copied_data.idx)
            copied_data.idx = np.array(range(len(copied_data.idx)))
            pipeline.last_idx_int = copied_data.idx[-1]
        return copied_data

    def convert_non_int_indexes_for_predict(self, pipeline):
        """Conversion non ``int`` (``datetime``, ``string``, etc) indexes in ``integer`` form on the predict stage
        """

        copied_data = deepcopy(self)
        is_timestamp = isinstance(copied_data.idx[0], pd._libs.tslibs.timestamps.Timestamp)
        is_numpy_datetime = isinstance(copied_data.idx[0], np.datetime64)
        # if predict stage - calculating shift from last train part index
        if is_timestamp or is_numpy_datetime:
            copied_data.supplementary_data.non_int_idx = copy(self.idx)
            copied_data.idx = self._resolve_non_int_idx(pipeline)
        elif not isinstance(copied_data.idx[0], (int, np.int32, np.int64)):
            # note, that string indexes do not have an order and always we think that indexes we want to predict go
            # immediately after the train indexes
            copied_data.supplementary_data.non_int_idx = copy(copied_data.idx)
            copied_data.idx = pipeline.last_idx_int + np.array(range(1, len(copied_data.idx) + 1))
        return copied_data

    @staticmethod
    def _resolve_func(pipeline, x):
        return pipeline.last_idx_int + (x - pipeline.last_idx_dt) // pipeline.period

    def _resolve_non_int_idx(self, pipeline):
        return np.array(list(map(lambda x: self._resolve_func(pipeline, x), self.idx)))


@dataclass
class OutputData(Data):
    """``Data`` type for data prediction in the node
    """

    features: Optional[np.ndarray] = None
    predict: np.ndarray = None
    target: Optional[np.ndarray] = None


def _resize_image(file_path: str, target_size: Tuple[int, int]):
    """Function resizes and rewrites the input image
    """

    img = cv2.imread(file_path)
    if img.shape[:2] != target_size:
        img = cv2.resize(img, (target_size[0], target_size[1]))
        cv2.imwrite(file_path, img)
    return img


def process_target_and_features(data_frame: pd.DataFrame,
                                target_column: Optional[Union[str, List[str]]]
                                ) -> Tuple[np.ndarray, Optional[np.ndarray]]:
    """Function process pandas ``dataframe`` with single column

    Args:
        data_frame: loaded pandas :obj:`DataFrame`
        target_column: names of columns with target or ``None``

    Returns:
        (``np.array`` (table) with features, ``np.array`` (column) with target)
    """

    if target_column == '':
        # Take the last column in the table
        target_column = data_frame.columns[-1]

    if target_column:
        target = atleast_2d(data_frame[target_column].to_numpy())
        features = data_frame.drop(columns=target_column).to_numpy()
    else:
        target = None
        features = data_frame.to_numpy()

    return features, target


def data_type_is_table(data: Union[InputData, OutputData]) -> bool:
    return data.data_type is DataTypesEnum.table


def data_type_is_ts(data: InputData) -> bool:
    return data.data_type is DataTypesEnum.ts


def data_type_is_multi_ts(data: InputData) -> bool:
    return data.data_type is DataTypesEnum.multi_ts


def data_type_is_text(data: InputData) -> bool:
    return data.data_type is DataTypesEnum.text


def data_type_is_image(data: InputData) -> bool:
    return data.data_type is DataTypesEnum.image


def get_indices_from_file(data_frame, file_path, idx_column='datetime'):
    if idx_column in data_frame.columns:
        df = pd.read_csv(file_path,
                         parse_dates=[idx_column])
        idx = [str(d) for d in df[idx_column]]
        return idx
    return np.arange(0, len(data_frame))


def array_to_input_data(features_array: np.array,
                        target_array: np.array,
                        idx: Optional[np.array] = None,
                        task: Task = Task(TaskTypesEnum.classification),
                        data_type: Optional[DataTypesEnum] = None):
    if idx is None:
        idx = np.arange(len(features_array))
    if data_type is None:
        data_type = autodetect_data_type(task)
    return InputData(idx=idx, features=features_array, target=target_array, task=task, data_type=data_type)


def autodetect_data_type(task: Task) -> DataTypesEnum:
    if task.task_type == TaskTypesEnum.ts_forecasting:
        return DataTypesEnum.ts
    else:
        return DataTypesEnum.table



import os
import sys
from subprocess import Popen
import subprocess
import pandas as pd

# for correct .bat files execution fedot should be build as a package in environment
env_name = sys.executable
env_path_placeholder = 'DEFAULT'
predictions_path = '../../fedot/api/predictions.csv'


def change_python_path(file_name, old, new):
    """ Function for changing env path in .bat for users settings"""
    with open(file_name, "r+") as file:
        text = file.read()
        file.seek(0)
        text = text.replace(old, new)
        file.write(text)
        file.truncate()


def run_console(bat_name):
    """ Function for running .bat files with returning prediction as df"""
    try:
        os.remove(predictions_path)
    except Exception:
        pass
    change_python_path(bat_name, env_path_placeholder, env_name)
    process = Popen(bat_name, creationflags=subprocess.CREATE_NEW_CONSOLE)
    process.wait()
    change_python_path(bat_name, env_name, env_path_placeholder)
    print(f"\nPrediction saved at {predictions_path}")
    df = pd.read_csv(predictions_path)
    return df


def run_cli_ts_forecasting():
    """ Test executing ts_forecasting problem from cli with saving prediction"""
    bat_name = 'cli_ts_call.bat'
    run_console(bat_name)


def run_cli_classification():
    """ Test executing classification problem from cli with saving prediction"""
    bat_name = 'cli_classification_call.bat'
    run_console(bat_name)


def run_cli_regression():
    """ Test executing regression problem from cli with saving prediction"""
    bat_name = 'cli_regression_call.bat'
    run_console(bat_name)


if __name__ == '__main__':
    run_cli_classification()



from typing import List, Iterable, Union

from fedot.core.data.data import OutputData, InputData
from fedot.core.data.merge.supplementary_data_merger import SupplementaryDataMerger
from fedot.core.log import default_log
from fedot.core.repository.dataset_types import DataTypesEnum
from fedot.core.data.array_utilities import *
from fedot.core.utilities.data_structures import are_same_length


class DataMerger:
    """
    Base class for merging a number of OutputData from one or several parent nodes
    into a single InputData for the next level node in the computation graph.

    Main method of the interface is `merge` that delegates merge stages to other methods.
    Merge process can be customized for data types by overriding these methods, primarily:
    `merge_targets`, `preprocess_predicts`, `merge_predicts`, `postprocess_predicts'.

    :param outputs: list with OutputData from parent nodes for merging
    """

    def __init__(self, outputs: List['OutputData'], data_type: DataTypesEnum = None):
        self.log = default_log(self)
        self.outputs = outputs
        self.data_type = data_type or DataMerger.get_datatype_for_merge(output.data_type for output in outputs)

        # Ensure outputs are of equal length, find common index if it is not
        idx_list = [np.asarray(output.idx) for output in outputs]
        self.common_indices = find_common_elements(*idx_list)
        if len(self.common_indices) == 0:
            raise ValueError(f'There are no common indices for outputs')

        # Find first output with the main target & resulting task
        self.main_output = DataMerger.find_main_output(outputs)

    @staticmethod
    def get(outputs: List['OutputData']) -> 'DataMerger':
        """ Construct appropriate data merger for the outputs. """

        # Ensure outputs can be merged
        data_type = DataMerger.get_datatype_for_merge(output.data_type for output in outputs)
        if data_type is None:
            raise ValueError("Can't merge different data types")

        merger_by_type = {
            DataTypesEnum.table: DataMerger,
            DataTypesEnum.ts: TSDataMerger,
            DataTypesEnum.multi_ts: TSDataMerger,
            DataTypesEnum.image: ImageDataMerger,
            DataTypesEnum.text: TextDataMerger,
        }
        cls = merger_by_type.get(data_type)
        if not cls:
            raise ValueError(f'Unable to merge data type {cls}')
        return cls(outputs, data_type)

    @staticmethod
    def get_datatype_for_merge(data_types: Iterable[DataTypesEnum]) -> Optional[DataTypesEnum]:
        # Check is all data types can be merged or not
        distinct = set(data_types)
        return distinct.pop() if len(distinct) == 1 else None

    def merge(self) -> 'InputData':
        common_idx = self.select_common(self.main_output.idx)

        filtered_main_target = self.merge_targets()

        common_predicts = self.find_common_predicts()
        mergeable_predicts = self.preprocess_predicts(common_predicts)
        merged_features = self.merge_predicts(mergeable_predicts)
        merged_features = self.postprocess_predicts(merged_features)

        updated_metadata = SupplementaryDataMerger(self.outputs, self.main_output).merge()

        return InputData(idx=common_idx, features=merged_features, target=filtered_main_target,
                         task=self.main_output.task, data_type=self.data_type,
                         supplementary_data=updated_metadata)

    def merge_targets(self) -> np.array:
        filtered_main_target = self.main_output.target
        # if target has the same form as index
        #  then it makes sense to extract target with common indices
        if filtered_main_target is not None and len(self.main_output.idx) == len(filtered_main_target):
            filtered_main_target = self.select_common(self.main_output.idx, filtered_main_target)
        return filtered_main_target

    def find_common_predicts(self) -> List[np.array]:
        """ Selects and returns only those elements of predicts that are common to all outputs. """

        # Forecast index is index with a length different from that of features/predictions.
        # Such index can't be used for extracting common predictions, and it must be
        # handled separately. This case arises for timeseries after lagged transform,
        # where the datatype becomes 'table', but we still must merge it as timeseries.
        is_forecast_indices = map(DataMerger.is_forecast_index, self.outputs)

        if any(is_forecast_indices):
            # Cut prediction length to minimum length
            predict_len = min(len(output.predict) for output in self.outputs)
            common_predicts = [output.predict[:predict_len] for output in self.outputs]
        else:
            common_predicts = [self.select_common(output.idx, output.predict) for output in self.outputs]
            if not are_same_length(common_predicts):
                raise ValueError('Indices of merged data are not equal and not unique. Check validity of the pipeline.')
        return common_predicts

    def preprocess_predicts(self, predicts: List[np.array]) -> List[np.array]:
        """ Pre-process (e.g. equalizes sizes, reshapes) and return list of arrays that can be merged. """
        return list(map(atleast_2d, predicts))

    def merge_predicts(self, predicts: List[np.array]) -> np.array:
        # Finally, merge predictions into features for the next stage
        return np.concatenate(predicts, axis=-1)

    def postprocess_predicts(self, merged_predicts: np.array) -> np.array:
        """ Post-process merged predictions (e.g. reshape). """
        return merged_predicts

    def select_common(self, idx: Union[list, np.array], data: Union[list, np.array] = None):
        """ Select elements from data according to index for data.
         Includes only elements with index from self.common_indices. """
        index_mask = np.isin(idx, self.common_indices)
        sliced = data if data is not None else idx
        sliced = np.asarray(sliced)[index_mask]
        return sliced

    @staticmethod
    def is_forecast_index(output: 'OutputData'):
        return len(output.idx) != len(output.predict)

    @staticmethod
    def find_main_output(outputs: List['OutputData']) -> 'OutputData':
        """ Returns first output with main target or (if there are
        no main targets) the output with priority secondary target. """
        priority_output = next((output for output in outputs
                                if output.supplementary_data.is_main_target), None)
        if not priority_output:
            flow_lengths = [output.supplementary_data.data_flow_length for output in outputs]
            i_priority_secondary = np.argmin(flow_lengths)
            priority_output = outputs[i_priority_secondary]
        return priority_output


class ImageDataMerger(DataMerger):

    def preprocess_predicts(self, predicts: List[np.array]) -> List[np.array]:
        # Reshape predicts to 4d (idx, width, height, channels)
        reshaped_predicts = list(map(atleast_4d, predicts))

        # And check image sizes
        img_wh = [predict.shape[1:3] for predict in reshaped_predicts]
        invalid_sizes = len(set(img_wh)) > 1  # Can merge only images of the same size
        if invalid_sizes:
            raise ValueError("Can't merge images of different sizes: " + str(img_wh))

        return reshaped_predicts


class TSDataMerger(DataMerger):

    def postprocess_predicts(self, merged_predicts: np.array) -> np.array:
        # Ensure that 1d-column timeseries remains 1d timeseries
        return flatten_extra_dim(merged_predicts)


class TextDataMerger(DataMerger):

    def merge_predicts(self, predicts: List[np.array]) -> np.array:
        if len(predicts) > 1:
            raise ValueError("Text tables and merge of text data is not supported")
        return predicts[0]

    def postprocess_predicts(self, merged_predicts: np.array) -> np.array:
        return merged_predicts
